# (c) 2012-2018 Dativa, all rights reserved
# -----------------------------------------
#  This code is licensed under MIT license (see license.txt for details)

import io
import re
import datetime
import pandas as pd
from pandas.api.types import is_numeric_dtype
from dativa.tools.pandas import get_column_name, get_unique_column_name, string_to_datetime, datetime_to_string, \
    is_numeric
from .base import BaseValidator, DateRangeType, HistoryType, SessionGapsType, SessionOverlapsType, ScrubberError, \
    ScrubberValidationError
from .fallback_addins import BestMatchesAddIn, FallbackAddIn, LookalikeAddIn, SkipBlankAddIn
from .anonymization import HashingAddIn, EncryptionAddIn, DecryptionAddIn


# noinspection PyUnresolvedReferences
class StringValidator(BaseValidator):
    """
    The string validator validates string columns in DataFrames
    """

    rule_type = "String"

    fields = {'minimum_length': {"default": None, "validator": lambda x: (0 <= x < 1e6)},
              'maximum_length': {"default": None, "validator": lambda x: (0 <= x < 1e6)},
              'regex': {"default": ".*", "validator": lambda x: re.compile(x, re.UNICODE) is not None},
              'is_unique': {"default": False, "validator": lambda x: x in [True, False]}}

    pre_execution = [DecryptionAddIn]
    post_execution = [SkipBlankAddIn, BestMatchesAddIn, LookalikeAddIn, FallbackAddIn, HashingAddIn,
                      EncryptionAddIn]

    def cross_validation(self):
        """
        Provides cross-validation on the fields passed to the string validator
        """
        if self.minimum_length > self.maximum_length:
            raise ScrubberValidationError(
                "minimum length cannot be greater than maximum length but {0} > {1}".format(self.minimum_length,
                                                                                            self.maximum_length))

    def _check_length(self, string):
        """
        Check that a string is within the defined length
        """
        if self.minimum_length <= len(string) <= self.maximum_length:
            return string
        else:
            return None

    def _run_string(self, string, regex):
        """
        Check the the string is valid
        """

        if pd.isnull(string):
            string = ""

        if isinstance(string, str):
            match = regex.match(string)
            if match:
                return self._check_length(string)
            else:
                return None
        else:
            return self._run_string(str(string), regex)

    def _get_regex(self):
        """
        compile a regex from the fields
        """
        if self.regex == '':
            return re.compile('.*', re.UNICODE)

        return re.compile(self.regex, re.UNICODE)

    def run(self, df, column):
        """
        Run the string validator against the passed column of the DataFrame
        """

        dirty_col = get_column_name(df, column)
        clean_col = get_unique_column_name(df, dirty_col)

        # create the regex
        regex = self._get_regex()

        if self.is_unique:
            dupes = df[df.duplicated(subset=dirty_col, keep='first')]
            if dupes.shape[0] > 0:
                self.report_writer.log_history(
                    self.rule_type,
                    dirty_col,
                    df[df.duplicated(subset=dirty_col, keep='first')],
                    HistoryType.REMOVED,
                    "Removed duplicates")
                df.drop_duplicates(
                    subset=dirty_col, keep='first', inplace=True)

        # add a new column for the clean data
        df[clean_col] = df[dirty_col].map(
            lambda word: self._run_string(word, regex))

        # set up the reference list for best matches...
        self.reference_list = pd.Series(df[clean_col].dropna().unique())


# noinspection PyUnresolvedReferences
class NumberValidator(BaseValidator):
    """
    The number validator runs validation against number fields in the data
    """

    rule_type = "Number"

    fields = {'decimal_places': {"default": None, "validator": lambda x: (0 <= x < 100)},
              'minimum_value': {"default": None, "validator": lambda x: (-1e24 <= x < 1e24)},
              'maximum_value': {"default": None, "validator": lambda x: (-1e24 <= x < 1e24)},
              'fix_decimal_places': {"default": None, "validator": lambda x: x in [True, False]},
              'is_unique': {"default": False, "validator": lambda x: x in [True, False]}}

    pre_execution = []
    post_execution = [SkipBlankAddIn, BestMatchesAddIn, LookalikeAddIn, FallbackAddIn]

    def cross_validation(self):
        """
        Provides cross-validation on the fields passed to the string validator
        """
        if self.minimum_value > self.maximum_value:
            raise ScrubberValidationError(
                "minimum value cannot be greater than maximum value but {0} > {1}".format(self.minimum_value,
                                                                                          self.maximum_value))

    def _get_regex(self):
        """
        Get the regex to validate for the number of decimal places
        """

        if self.decimal_places == 0:
            regex_string = "^\\d+$"
        else:
            regex_string = "^-?\\d+\\." + "\\d" * self.decimal_places + "$"
        return re.compile(regex_string, re.UNICODE)

    @staticmethod
    def _force_format_string(number_string, format_string):
        """
        Forces the number to the specified format string
        """
        return format_string.format(float(number_string))

    def _validate_number(self, string, regex, format_string):
        """
        Run number validation
        """

        # first check that it is in range
        try:
            number = float(string)
            if not self.minimum_value <= number <= self.maximum_value:
                return None
        except ValueError:
            return None

        # now check the decimal places...
        if self.fix_decimal_places:
            # attempt to fix dirty data with correct decimcal places
            return self._force_format_string(string, format_string)
        else:
            match = regex.match(string)
            if match:
                return string
            else:
                return None

    def run(self, df, column):
        """
        Run the number validator against the passed column of the DataFrame
        """

        dirty_col = get_column_name(df, column)
        clean_col = get_unique_column_name(df, dirty_col)
        write_report_dp = None

        # check range then (optionally) fix decimal places
        if is_numeric_dtype(df[dirty_col]):
            # we are validating a number
            df[clean_col] = df.loc[
                (self.minimum_value <= df[dirty_col]) & (df[dirty_col] <= self.maximum_value), dirty_col].round(
                self.decimal_places)

            if not self.fix_decimal_places:
                df.loc[df[clean_col] != df[dirty_col], clean_col] = None
            elif self.fix_decimal_places:

                # all entries which are rounded (regardless of if they matched already)
                write_report_dp = df.loc[
                    (df[clean_col] == df[dirty_col].round(self.decimal_places)), [dirty_col, clean_col]]
        else:
            # we are validating a string
            regex = self._get_regex()
            format_string = "{{0:.{0}f}}".format(self.decimal_places)

            df[clean_col] = df[dirty_col].map(lambda word: self._validate_number(word, regex, format_string))

            if self.fix_decimal_places and (df[df[clean_col] != df[dirty_col]].dropna().shape[0] > 0):
                write_report_dp = df[df[clean_col] != df[dirty_col]].dropna()

        # write report if it has changed the number of decimal places
        if write_report_dp is not None:
            self.report_writer.log_history(
                self.rule_type,
                dirty_col,
                write_report_dp,
                HistoryType.MODIFIED,
                "Automatically fixed to {0} decimal places".format(self.decimal_places))

        # remove duplicated entries
        if self.is_unique:
            dupes = df[df.duplicated(subset=clean_col, keep='first')]
            if dupes.shape[0] > 0:
                self.report_writer.log_history(
                    self.rule_type,
                    clean_col,
                    dupes,
                    HistoryType.REMOVED,
                    "Removed duplicates")
                df.drop_duplicates(
                    subset=clean_col, keep='first', inplace=True)

        # set up the reference list for best matches...
        self.reference_list = pd.Series(df[clean_col].dropna().unique())


# noinspection PyUnresolvedReferences
class DateValidator(BaseValidator):
    """
    Runs data validation
    """

    rule_type = "Date"

    fields = {'date_format': {"default": None,
                              "validator": lambda x: datetime_to_string(datetime.datetime.now(), x) is not None},
              'range_check': {"default": DateRangeType.NONE,
                              "validator": lambda x: x in [DateRangeType.NONE, DateRangeType.FIXED,
                                                           DateRangeType.ROLLING]},
              'range_minimum': {"default": "none", "validator": lambda x: True},
              'range_maximum': {"default": "none", "validator": lambda x: True},
              'is_unique': {"default": False, "validator": lambda x: x in [True, False]}}

    pre_execution = []
    post_execution = [SkipBlankAddIn, BestMatchesAddIn, LookalikeAddIn, FallbackAddIn]

    def cross_validation(self):
        """
        Provides cross-validation on the fields passed to the date validator
        """
        if self.range_check != DateRangeType.NONE:
            if self.range_minimum == "none":
                raise ScrubberValidationError(
                    "range_minimum must be set if range_check is not set to {0}".format(DateRangeType.NONE))
            if self.range_maximum == "none":
                raise ScrubberValidationError(
                    "range_maximum must be set if range_check is not set to {0}".format(DateRangeType.NONE))
            if self.range_check == DateRangeType.ROLLING:
                if not is_numeric(self.range_minimum):
                    raise ScrubberValidationError(
                        "range_minimum must be a number if range_type={0}".format(DateRangeType.ROLLING))
                if not is_numeric(self.range_maximum):
                    raise ScrubberValidationError(
                        "range_minimum must be a number if range_type={0}".format(DateRangeType.ROLLING))
            else:
                if string_to_datetime(self.range_minimum, self.date_format, errors="coerce") is pd.NaT:
                    raise ScrubberValidationError(
                        "range_minimum does not match the data format {0}".format(self.date_format))
                if string_to_datetime(self.range_maximum, self.date_format, errors="coerce") is pd.NaT:
                    raise ScrubberValidationError(
                        "range_maximum does not match the data format {0}".format(self.date_format))
            if self.range_minimum > self.range_maximum:
                raise ScrubberValidationError(
                    "range_minimum cannot be greater than range_maximum but {0} > {1}".format(self.range_minimum,
                                                                                              self.range_maximum))

    def run(self, df, column):
        """
        Run the date validator against the passed column of the DataFrame
        """
        dirty_col = get_column_name(df, column)
        clean_col = get_unique_column_name(df, dirty_col)
        date_col = get_unique_column_name(df, "zxc34324uih")

        if self.is_unique:
            dupes = df[df.duplicated(subset=dirty_col, keep='first')]
            if dupes.shape[0] > 0:
                self.report_writer.log_history(
                    self.rule_type,
                    dirty_col,
                    dupes,
                    HistoryType.REMOVED,
                    "Removed duplicates")
                df.drop_duplicates(
                    subset=dirty_col, keep='first', inplace=True)

        if pd.core.dtypes.common.is_datetime_or_timedelta_dtype(df[dirty_col]):
            df[date_col] = df[dirty_col]
        else:
            df[date_col] = string_to_datetime(df[dirty_col], format=self.date_format, errors='coerce')
        if self.range_check != DateRangeType.NONE:
            if self.range_check == DateRangeType.ROLLING:
                min_date = datetime.datetime.now() + datetime.timedelta(days=int(self.range_minimum))
                max_date = datetime.datetime.now() + datetime.timedelta(days=int(self.range_maximum))
            else:
                min_date = string_to_datetime(
                    self.range_minimum, self.date_format)
                max_date = string_to_datetime(
                    self.range_maximum, self.date_format)

            df.loc[df[date_col] < min_date, date_col] = pd.np.NaN
            df.loc[df[date_col] > max_date, date_col] = pd.np.NaN

        # now update the clean column...
        df.loc[df[date_col].notnull(), clean_col] = datetime_to_string(
            df[date_col], self.date_format)
        df.drop(date_col, axis=1, inplace=True)

        # set up the reference list for best matches...
        self.reference_list = pd.Series(df[clean_col].dropna().unique())


# noinspection PyUnresolvedReferences
class LookupValidator(BaseValidator):
    """
    """

    rule_type = "Lookup"

    fields = {'original_reference': {"default": None, "validator": lambda x: True},
              'reference_field': {"default": None, "validator": lambda x: True},
              'blacklist': {"default": False, "validator": lambda x: x in (True, False)}}

    pre_execution = []
    post_execution = [SkipBlankAddIn, BestMatchesAddIn, LookalikeAddIn, FallbackAddIn]

    def _get_reference_df(self):
        """
        Loads the reference data into a DataFrame
        """

        # load the initially provided data
        reference_words = self.get_df(self.original_reference)
        reference_column = get_column_name(reference_words, self.reference_field)

        # drop the other columns
        if reference_words.shape[1] > 1:
            for col in reference_words.columns:
                if col != reference_column:
                    reference_words.drop(col, axis=1, inplace=True)

        reference_words.drop_duplicates(
            subset=reference_column, keep='first', inplace=True)

        return reference_words, reference_column

    def run(self, df, column):
        """
        Run the lookup validator against the passed column of the DataFrame
        """

        reference_df, reference_column = self._get_reference_df()

        dirty_col = get_column_name(df, column)
        clean_col = get_unique_column_name(df, dirty_col)
        ref_col = reference_column

        # join the dirty DataFrame to the clean DataFrame for any exact matches
        reference_df.rename(
            columns={ref_col: clean_col}, inplace=True)
        cleaned = df.loc[:, [dirty_col]].merge(
            reference_df, left_on=dirty_col, right_on=clean_col, copy=False, how='left').set_index(df.index)

        if self.blacklist:
            # exclude any matches...
            df[clean_col] = df.loc[cleaned[clean_col].isnull(), dirty_col]

            # setup the reference list for the best matches...
            self.reference_list = pd.Series(df[clean_col].dropna().unique())
        else:
            # set any matching columns
            df[clean_col] = cleaned[clean_col]

            # use the reference dataframe as the reference list for best matches
            self.reference_list = reference_df[clean_col].copy()


# noinspection PyUnresolvedReferences
class UniqueFieldsValidator(BaseValidator):
    """
    Tests combinations of columns for uniqueness
    """

    rule_type = "Uniqueness"

    fields = {'use_last_value': {"default": False, "validator": lambda x: x in [True, False]},
              'unique_fields': {"default": None, "validator": lambda x: True}}

    pre_execution = []
    post_execution = []

    def _get_columns_from_unique(self, df, separator=","):
        """
        Get a validated list of columns based on the passed string of unique fields
        """
        unique_fields = self.unique_fields.split(separator)
        cols = df.columns.values.tolist()

        # get the unique fields...
        for field in unique_fields:
            if field not in cols:
                raise ScrubberError(
                    "The field {0} was not present in the source file".format(field))

        return unique_fields

    def _get_keep_option(self):
        """
        Returns the pandas keep option based on whether to keep the first
        or last value
        """

        if self.use_last_value is True:
            keep_option = 'last'
            return keep_option
        else:
            keep_option = 'first'
            return keep_option

    def run(self, df):
        """
        Run the automatic uniqueness validator algorithm on the passed columns of the DataFrame
        """

        keep_option = self._get_keep_option()
        unique_columns = self._get_columns_from_unique(df)

        dupes = df[df.duplicated(subset=unique_columns, keep=keep_option)]
        if dupes.shape[0] > 0:
            self.report_writer.log_history(
                self.rule_type,
                self.unique_fields,
                df[df.duplicated(subset=unique_columns, keep=keep_option)],
                HistoryType.REMOVED,
                "Removed duplicates")
            df.drop_duplicates(subset=unique_columns,
                               keep=keep_option, inplace=True)


# noinspection PyUnresolvedReferences
class SessionValidator(BaseValidator):
    """
    Runs session validation
    """

    rule_type = "Session"

    fields = {'key_field': {"default": None, "validator": lambda x: True},
              'start_field': {"default": None, "validator": lambda x: True},
              'end_field': {"default": None, "validator": lambda x: True},
              'overlaps_option': {"default": None, "validator": lambda x: x in [SessionOverlapsType.IGNORE,
                                                                                SessionOverlapsType.TRUNCATE_START,
                                                                                SessionOverlapsType.TRUNCATE_END]},
              'gaps_option': {"default": None,
                              "validator": lambda x: x in [SessionGapsType.IGNORE, SessionGapsType.EXTEND_START,
                                                           SessionGapsType.EXTEND_END, SessionGapsType.INSERT_NEW]},
              'date_format': {"default": None,
                              "validator": lambda x: datetime_to_string(datetime.datetime.now(), x) is not None},
              'allowed_gap_seconds': {"default": 0, "validator": lambda x: (0 <= x <= 3600)},
              'allowed_overlap_seconds': {"default": 0, "validator": lambda x: (0 <= x <= 3600)},
              'template_for_new': {"default": "", "validator": lambda x: True},
              'remove_zero_length': {"default": True, "validator": lambda x: True}}

    pre_execution = []
    post_execution = []

    @staticmethod
    def _string_to_df(string, columns):
        """
        Returns a DataFrame from a passed string
        """
        f = io.StringIO()
        f.write(string)
        f.seek(0)
        return pd.read_csv(f,
                           dtype=pd.np.str,
                           skipinitialspace=True,
                           encoding="UTF-8",
                           delimiter=",",
                           quotechar="\"",
                           header=None,
                           names=columns)

    def _setup_columns(self, df):
        """
        Sets up the additional columns needed in the DataFrame and the columns dictionary
        """
        cols = dict()
        cols["key"] = get_column_name(df, self.key_field)
        cols["next_key"] = get_unique_column_name(df, "next_key")
        cols["previous_key"] = get_unique_column_name(df, "previous_key")
        cols["start"] = get_unique_column_name(df, "start")
        cols["end"] = get_unique_column_name(df, "end")
        cols["next_start"] = get_unique_column_name(df, "next")
        cols["previous_end"] = get_unique_column_name(df, "previous")

        cols["original_start"] = get_column_name(df, self.start_field)
        cols["original_end"] = get_column_name(df, self.end_field)

        df[cols["start"]] = string_to_datetime(
            df[cols["original_start"]], format=self.date_format)
        df[cols["end"]] = string_to_datetime(
            df[cols["original_end"]], format=self.date_format)

        df.sort_values([cols["key"], cols["start"], cols["end"]], inplace=True)

        df[cols["next_key"]] = df[cols["key"]].shift(-1)
        df[cols["previous_key"]] = df[cols["key"]].shift(1)
        df[cols["next_start"]] = df[cols["start"]].shift(-1)
        df[cols["previous_end"]] = df[cols["end"]].shift(1)

        cols["clean_start"] = get_unique_column_name(
            df, get_column_name(df, self.start_field))
        cols["clean_end"] = get_unique_column_name(
            df, get_column_name(df, self.end_field))

        return cols

    @staticmethod
    def _clean_up_columns(df, cols):
        """
        Cleans up the additional columns in the DataFrame
        """
        df.drop(cols["next_key"], axis=1, inplace=True)
        df.drop(cols["previous_key"], axis=1, inplace=True)
        df.drop(cols["next_start"], axis=1, inplace=True)
        df.drop(cols["previous_end"], axis=1, inplace=True)
        df.drop(cols["start"], axis=1, inplace=True)
        df.drop(cols["end"], axis=1, inplace=True)
        if cols["clean_start"] in df.columns:
            df.drop(cols["clean_start"], axis=1, inplace=True)
        if cols["clean_end"] in df.columns:
            df.drop(cols["clean_end"], axis=1, inplace=True)

    def _get_rows_to_truncate(self, df, cols):
        """
        Returns a set of rows with overlapping sessions
        """
        number_seconds = pd.Timedelta(datetime.timedelta(
            seconds=self.allowed_overlap_seconds))
        if self.overlaps_option == SessionOverlapsType.TRUNCATE_START:
            return (df[cols["previous_end"]] - df[cols["start"]] > number_seconds) & (
                    df[cols["key"]] == df[cols["previous_key"]])
        else:
            return (df[cols["end"]] - df[cols["next_start"]] > number_seconds) & (
                    df[cols["key"]] == df[cols["next_key"]])

    def _run_truncation(self, df, cols):
        """
        Truncates overlapping sessions
        """
        rows = self._get_rows_to_truncate(df, cols)
        if self.overlaps_option == SessionOverlapsType.TRUNCATE_START:
            if rows[rows].shape[0] > 0:
                df.loc[rows, cols["clean_start"]] = df[cols["previous_end"]]
                self.report_writer.log_history(
                    self.rule_type,
                    cols["original_start"],
                    df.loc[rows, [cols["key"], cols["original_start"],
                                  cols["original_end"], cols["clean_start"]]],
                    HistoryType.MODIFIED,
                    "Truncated start")
        else:
            if rows[rows].shape[0] > 0:
                df.loc[rows, cols["clean_end"]] = df[cols["next_start"]]
                self.report_writer.log_history(
                    self.rule_type,
                    cols["original_end"],
                    df.loc[rows, [cols["key"], cols["original_start"],
                                  cols["original_end"], cols["clean_end"]]],
                    HistoryType.MODIFIED,
                    "Truncated end")

    def _get_rows_to_extend(self, df, cols):
        """
        Gets a list of rows with gaps
        """
        number_seconds = pd.Timedelta(
            datetime.timedelta(seconds=self.allowed_gap_seconds))
        if self.gaps_option == SessionGapsType.EXTEND_START:
            return (df[cols["start"]] - df[cols["previous_end"]] > number_seconds) & (
                    df[cols["key"]] == df[cols["previous_key"]])
        else:
            return (df[cols["next_start"]] - df[cols["end"]] > number_seconds) & (
                    df[cols["key"]] == df[cols["next_key"]])

    def _run_extension(self, df, cols):
        """
        Extends sessions to fill gaps
        """
        rows = self._get_rows_to_extend(df, cols)
        if self.gaps_option == SessionGapsType.EXTEND_START:
            if rows.shape[0] > 0:
                df.loc[rows, cols["clean_start"]] = df[cols["previous_end"]]
                self.report_writer.log_history(
                    self.rule_type,
                    cols["original_start"],
                    df.loc[rows, [cols["key"], cols["original_start"],
                                  cols["original_end"], cols["clean_start"]]],
                    HistoryType.MODIFIED,
                    "Extended start")
        else:
            if rows.shape[0] > 0:
                df.loc[rows, cols["clean_end"]] = df[cols["next_start"]]
                self.report_writer.log_history(
                    self.rule_type,
                    cols["original_end"],
                    df.loc[rows, [cols["key"], cols["original_start"],
                                  cols["original_end"], cols["clean_end"]]],
                    HistoryType.MODIFIED,
                    "Extended end")

    def _fill_gaps(self, df, cols, template_row):
        """
        Fills gaps in the session with new records
        """
        number_seconds = pd.Timedelta(
            datetime.timedelta(seconds=self.allowed_gap_seconds))
        new_rows = df[((df[cols["next_start"]] - df[cols["end"]] > number_seconds) &
                       (df[cols["key"]] == df[cols["next_key"]]))].copy()

        if new_rows.shape[0] > 0:
            new_rows[cols["original_start"]] = new_rows[cols["original_end"]]
            new_rows[cols["original_end"]] = new_rows[cols["next_start"]].apply(
                lambda x: datetime_to_string(x, self.date_format))

            for col in template_row.columns:
                if col not in [cols["original_start"], cols["original_end"], cols["key"]]:
                    new_rows[col] = template_row[col][0]

            for i, row in new_rows.iterrows():
                df.loc[len(df)] = row

            df.sort_values([cols["key"], cols["original_start"],
                            cols["original_end"]], inplace=True)

            df[cols["start"]] = string_to_datetime(
                df[cols["original_start"]], format=self.date_format)
            df[cols["end"]] = string_to_datetime(
                df[cols["original_end"]], format=self.date_format)

            self.report_writer.log_history(
                self.rule_type,
                cols["original_start"],
                new_rows[[cols["key"], cols["original_start"],
                          cols["original_end"]]],
                HistoryType.INSERTED,
                'Filled gaps')

    def _remove_zero_length_sessions(self, df, cols):
        """
        Deletes any null sessions from the dataframe
        """
        if cols["clean_start"] in df.columns:
            rows = df[df[cols["end"]] == df[cols["clean_start"]]]
        elif cols["clean_end"] in df.columns:
            rows = df[df[cols["start"]] == df[cols["clean_end"]]]
        else:
            rows = df[df[cols["start"]] == df[cols["end"]]]

        if rows.shape[0] > 0:
            self.report_writer.log_history(
                self.rule_type,
                cols["original_start"],
                rows,
                HistoryType.REMOVED,
                'Removed zero length')

            df.drop(rows.index, inplace=True)

    def run(self, df):
        """
        Run the automatic session validator algorithm on the passed column of the DataFrame
        """
        cols = self._setup_columns(df)

        if self.overlaps_option in [SessionOverlapsType.TRUNCATE_START, SessionOverlapsType.TRUNCATE_END]:
            self._run_truncation(df, cols)

        if self.gaps_option in [SessionGapsType.EXTEND_START, SessionGapsType.EXTEND_END]:
            self._run_extension(df, cols)

        if self.gaps_option == SessionGapsType.INSERT_NEW and self.template_for_new != '':
            template_row = self._string_to_df(
                self.template_for_new, df.columns)
            self._fill_gaps(df, cols, template_row)

        if self.remove_zero_length:
            self._remove_zero_length_sessions(df, cols)

        if cols["clean_start"] in df.columns:
            df.loc[df[cols["clean_start"]].notnull(), cols["original_start"]] = df.loc[df[cols["clean_start"]].notnull(
            ), cols["clean_start"]].apply(lambda x: datetime_to_string(x, self.date_format))

        if cols["clean_end"] in df.columns:
            df.loc[df[cols["clean_end"]].notnull(), cols["original_end"]] = df.loc[df[cols["clean_end"]].notnull(
            ), cols["clean_end"]].apply(lambda x: datetime_to_string(x, self.date_format))

        self._clean_up_columns(df, cols)

    def count_good_rows(self, df):
        """
        Checks for gaps and overlaps without altering the DataFrame.
        Returns the number of good rows in the DatFrame
        """

        cols = self._setup_columns(df)

        exact_matches = df[(df[cols["next_start"]] == df[cols["end"]]) & (
                df[cols["key"]] == df[cols["next_key"]])].shape[0]
        ineligible = df[(df[cols["key"]] != df[cols["next_key"]])].shape[0]

        gaps = df[(df[cols["next_start"]] > df[cols["end"]]) & (
                df[cols["key"]] == df[cols["next_key"]])].shape[0]
        overlaps = df[(df[cols["next_start"]] < df[cols["end"]]) & (
                df[cols["key"]] == df[cols["next_key"]])].shape[0]

        self._clean_up_columns(df, cols)

        number_good_rows = exact_matches
        if self.gaps_option == SessionGapsType.IGNORE:
            number_good_rows = number_good_rows + gaps
        if self.overlaps_option == SessionOverlapsType.IGNORE:
            number_good_rows = number_good_rows + overlaps

        return number_good_rows + ineligible



# (c) 2012-2018 Dativa, all rights reserved
# -----------------------------------------
# Usage subject to license agreement
# hello@dativa.com for more information

from dativa.tools.pandas import CSVHandler
from dativa.scrubber.tests import _BaseTest
from dativa.scrubber import Scrubber, ScrubberValidationError


class LookupTests(_BaseTest):

    def test_blacklist(self):
        self._test_filter(dirty_file="lookup/test_cities_dirty.csv",
                          clean_file="lookup/test_cities_dirty_blacklist.csv",
                          config={
                              "rules": [
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": True,
                                          "blacklist": True,
                                          "fallback_mode": "remove_record",
                                          "original_reference": "lookup/test_cities_reference.csv",
                                          "reference_field": 0
                                      },
                                      "rule_type": "Lookup",
                                      "field": "1"
                                  }
                              ]
                          },
                          csv_header=-1,
                          report=[
                              {

                                  'field': 1,
                                  'rule': 'Lookup',
                                  'number_records': 64,
                                  'category': 'replaced',
                                  'description': 'Best match',
                                  'df': [['Bristol', 'Bsitrol'],
                                         ['London', 'Lndon'],
                                         ['Bristol', 'Bsitrol'],
                                         ['Bristol', 'Bsitrol'],
                                         ['Bristol', 'Bsitrol'],
                                         ['Bristol', 'Bsitrol'],
                                         ['Bristol', 'Bsitrol'],
                                         ['Bristol', 'Bsitrol'],
                                         ['Bristol', 'Bsitrol'],
                                         ['Bristol', 'Bsitrol'],
                                         ['Bristol', 'Bsitrol'],
                                         ['Bristol', 'Bsitrol'],
                                         ['Bristol', 'Bsitrol'],
                                         ['London', 'Lndon'],
                                         ['London', 'Lndon'],
                                         ['London', 'Lndon'],
                                         ['London', 'Lndon'],
                                         ['London', 'Lndon'],
                                         ['London', 'Lndon'],
                                         ['London', 'Lndon'],
                                         ['London', 'Lndon'],
                                         ['London', 'Lndon'],
                                         ['London', 'Lndon'],
                                         ['London', 'Lndon'],
                                         ['London', 'Lndon'],
                                         ['London', 'Lndon'],
                                         ['London', 'Lndon'],
                                         ['London', 'Lndon'],
                                         ['London', 'Lndon'],
                                         ['London', 'Lndon'],
                                         ['London', 'Lndon'],
                                         ['London', 'Lndon'],
                                         ['London', 'Lndon'],
                                         ['London', 'Lndon'],
                                         ['Liverpool', 'Lverpool'],
                                         ['Liverpool', 'Lverpool'],
                                         ['Liverpool', 'Lverpool'],
                                         ['Liverpool', 'Lverpool'],
                                         ['Liverpool', 'Lverpool'],
                                         ['Liverpool', 'Lverpool'],
                                         ['Liverpool', 'Lverpool'],
                                         ['Liverpool', 'Lverpool'],
                                         ['Liverpool', 'Lverpool'],
                                         ['Liverpool', 'Lverpool'],
                                         ['Liverpool', 'Lverpool'],
                                         ['Liverpool', 'Lverpool'],
                                         ['Liverpool', 'Lverpool'],
                                         ['Liverpool', 'Lverpool'],
                                         ['Liverpool', 'Lverpool'],
                                         ['Liverpool', 'Lverpool'],
                                         ['Liverpool', 'Lverpool'],
                                         ['Liverpool', 'Lverpool'],
                                         ['Liverpool', 'Lverpool'],
                                         ['Liverpool', 'Lverpool'],
                                         ['Liverpool', 'Lverpool'],
                                         ['Liverpool', 'Lverpool'],
                                         ['Liverpool', 'Lverpool'],
                                         ['Liverpool', 'Lverpool'],
                                         ['Liverpool', 'Lverpool'],
                                         ['Liverpool', 'Lverpool'],
                                         ['Liverpool', 'Lverpool'],
                                         ['Liverpool', 'Lverpool'],
                                         ['Liverpool', 'Lverpool'],
                                         ['Liverpool', 'Lverpool']],

                              },
                          ])

    def test_lookup(self):
        self._test_filter(dirty_file="lookup/test_cities_dirty.csv",
                          clean_file="lookup/test_cities_dirty_no_banana.csv",
                          config={
                              "rules": [
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": True,
                                          "default_value": "Banana",
                                          "fallback_mode": "do_not_replace",
                                          "lookalike_match": False,
                                          "original_reference": "lookup/test_cities_reference.csv",

                                          "reference_field": "-1",
                                          "skip_blank": False,
                                          "string_distance_threshold": 0.7
                                      },
                                      "rule_type": "Lookup",
                                      "field": "1"
                                  }
                              ]
                          },
                          report=[
                              {

                                  'field': 'city',
                                  'rule': 'Lookup',
                                  'number_records': 25,
                                  'category': 'replaced',
                                  'description': 'Best match',
                                  'df': [['Bsitrol', 'Bristol'],
                                         ['London', 'Croydon'],
                                         ['Lverpool', 'Liverpool'],
                                         ['méémlent', 'élément'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon']],

                              },
                              {

                                  'field': 'city',
                                  'rule': 'Lookup',
                                  'number_records': 3,
                                  'category': 'ignored',
                                  'description': 'No changes made',
                                  'df': [['Bath', 'Bath'],
                                         ['Btha', 'Btha'],
                                         ['Lndon', 'Lndon']],

                              },
                          ])

    def test_lookup_1(self):
        self._test_filter(dirty_file="lookup/test_cities_dirty.csv",
                          clean_file="lookup/test_cities_dirty_banana.csv",
                          config={
                              "rules": [
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": True,
                                          "default_value": "Banana",
                                          "fallback_mode": "use_default",
                                          "lookalike_match": False,
                                          "original_reference": "lookup/test_cities_reference.csv",

                                          "reference_field": "-1",
                                          "skip_blank": False,
                                          "string_distance_threshold": 0.7
                                      },
                                      "rule_type": "Lookup",
                                      "field": "1"
                                  }
                              ]
                          },
                          report=[
                              {

                                  'field': 'city',
                                  'rule': 'Lookup',
                                  'number_records': 25,
                                  'category': 'replaced',
                                  'description': 'Best match',
                                  'df': [['Bsitrol', 'Bristol'],
                                         ['London', 'Croydon'],
                                         ['Lverpool', 'Liverpool'],
                                         ['méémlent', 'élément'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon']],

                              },
                              {

                                  'field': 'city',
                                  'rule': 'Lookup',
                                  'number_records': 3,
                                  'category': 'replaced',
                                  'description': 'Replaced with default value',
                                  'df': [['Bath', 'Banana'],
                                         ['Btha', 'Banana'],
                                         ['Lndon', 'Banana']],

                              },
                          ])

    def test_lookup_2(self):
        self._test_filter(dirty_file="lookup/test_cities_dirty.csv",
                          clean_file="lookup/test_cities_dirty_deleted_banana.csv",
                          config={
                              "rules": [
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": True,
                                          "default_value": "Banana",
                                          "fallback_mode": "remove_record",
                                          "lookalike_match": False,
                                          "original_reference": "lookup/test_cities_reference.csv",

                                          "reference_field": "-1",
                                          "skip_blank": False,
                                          "string_distance_threshold": 0.7
                                      },
                                      "rule_type": "Lookup",
                                      "field": "1"
                                  }
                              ]
                          },
                          report=[
                              {

                                  'field': 'city',
                                  'rule': 'Lookup',
                                  'number_records': 25,
                                  'category': 'replaced',
                                  'description': 'Best match',
                                  'df': [['Bsitrol', 'Bristol'],
                                         ['London', 'Croydon'],
                                         ['Lverpool', 'Liverpool'],
                                         ['méémlent', 'élément'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon']],

                              },
                              {

                                  'field': 'city',
                                  'rule': 'Lookup',
                                  'number_records': 3,
                                  'category': 'quarantined',
                                  'description': 'Data quarantined',
                                  'df': [['Bath'],
                                         ['Btha'],
                                         ['Lndon']],

                              },
                          ])

    def test_lookup_short(self):
        self._test_filter(dirty_file="lookup/test_short_original.csv",
                          clean_file="lookup/test_short_cleaned.csv",
                          config={
                              "rules": [
                                  {"field": "word",
                                   "append_results": False,
                                   "rule_type": "Lookup",
                                   "params": {
                                       "fallback_mode": "remove_record",
                                       "reference_field": "word",
                                       "original_reference": "lookup/test_short_reference.csv",
                                   }
                                   }
                              ]
                          },
                          report=[
                              {

                                  'field': 'word',
                                  'rule': 'Lookup',
                                  'number_records': 1,
                                  'category': 'quarantined',
                                  'description': 'Data quarantined',
                                  'df': [['E']],

                              },
                          ])

    def test_lookup_short_bad_field(self):
        self._test_filter(dirty_file="lookup/test_short_original.csv",
                          clean_file="lookup/test_short_cleaned.csv",
                          config={
                              "rules": [
                                  {"field": "word",
                                   "append_results": False,
                                   "rule_type": "Lookup",
                                   "params": {
                                       "fallback_mode": "remove_record",
                                       "reference_field": "word_not_there",
                                       "original_reference": "lookup/test_short_reference.csv",
                                   }
                                   }
                              ]
                          },
                          expected_error=KeyError)

    def test_df_dict_error(self):
        csv = CSVHandler(base_path=self.base_path)

        df = csv.get_dataframe("lookup/test_short_original.csv")

        scrubber = Scrubber()

        try:
            scrubber.run(df=df,
                         config={
                             "rules": [
                                 {"field": "word",
                                  "append_results": False,
                                  "rule_type": "Lookup",
                                  "params": {
                                      "fallback_mode": "remove_record",
                                      "reference_field": "word",
                                      "original_reference": "lookup/test_short_reference.csv",
                                  }
                                  }
                             ]
                         }, df_dict=[])
        except ScrubberValidationError:
            self.assertTrue(True)

    def test_duplicates_in_lookup(self):

        self._test_filter(dirty_file="lookup/test_duplicates_original.csv",
                          clean_file="lookup/test_duplicates_cleaned.csv",
                          config={
                              "rules": [
                                  {"field": "word",
                                   "append_results": False,
                                   "rule_type": "Lookup",
                                   "params": {
                                       "fallback_mode": "remove_record",
                                       "reference_field": "word",
                                       "original_reference": "lookup/test_duplicates_reference.csv",
                                   }
                                   }
                              ]
                          },
                          report=[
                              {

                                  'field': 'word',
                                  'rule': 'Lookup',
                                  'number_records': 1,
                                  'category': 'quarantined',
                                  'description': 'Data quarantined',
                                  'df': [['E']],

                              },
                          ])




# (c) 2012-2018 Dativa, all rights reserved
# -----------------------------------------
# Usage subject to license agreement
# hello@dativa.com for more information

from dativa.scrubber.tests import _BaseTest


class BestMatchTests(_BaseTest):

    def test_best_match(self):
        self._test_filter(dirty_file="generic/US1_movies_dirty.csv",
                          clean_file="generic/US1_movies_clean.csv",
                          config={
                              "rules": [
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": True,
                                          "default_value": "N/A",
                                          "fallback_mode": "do_not_replace",
                                          "lookalike_match": False,
                                          "original_reference": "generic/TV_Titles_Master_50000_records.csv",

                                          "reference_field": "0",
                                          "skip_blank": False,
                                          "string_distance_threshold": 0.7
                                      },
                                      "rule_type": "Lookup",
                                      "field": "0"
                                  },
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": True,
                                          "default_value": "N/A",
                                          "fallback_mode": "do_not_replace",
                                          "lookalike_match": False,
                                          "original_reference": "generic/TV_Titles_Master_50000_records.csv",

                                          "reference_field": "0",
                                          "skip_blank": False,
                                          "string_distance_threshold": 0.7
                                      },
                                      "rule_type": "Lookup",
                                      "field": "1"
                                  },
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": True,
                                          "default_value": "N/A",
                                          "fallback_mode": "use_default",
                                          "is_unique": False,
                                          "lookalike_match": False,
                                          "maximum_length": 1000,
                                          "minimum_length": 0,
                                          "regex": "^0$",
                                          "skip_blank": False,
                                          "string_distance_threshold": 0.7
                                      },
                                      "rule_type": "String",
                                      "field": "2"
                                  },
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": True,
                                          "default_value": "N/A",
                                          "fallback_mode": "use_default",
                                          "is_unique": False,
                                          "lookalike_match": False,
                                          "maximum_length": 1000,
                                          "minimum_length": 0,
                                          "regex": ".+",
                                          "skip_blank": False,
                                          "string_distance_threshold": 0.7
                                      },
                                      "rule_type": "String",
                                      "field": "3"
                                  },
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": True,
                                          "default_value": "BaNonea",
                                          "fallback_mode": "do_not_replace",
                                          "lookalike_match": False,
                                          "original_reference": "generic/Genres_Master.csv",

                                          "reference_field": "0",
                                          "skip_blank": False,
                                          "string_distance_threshold": 0.7
                                      },
                                      "rule_type": "Lookup",
                                      "field": "7"
                                  },
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": True,
                                          "default_value": "BaNonea",
                                          "fallback_mode": "do_not_replace",
                                          "lookalike_match": False,
                                          "original_reference": "generic/Genres_Master.csv",

                                          "reference_field": "0",
                                          "skip_blank": False,
                                          "string_distance_threshold": 0.7
                                      },
                                      "rule_type": "Lookup",
                                      "field": "8"
                                  },
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": True,
                                          "default_value": "BaNonea",
                                          "fallback_mode": "do_not_replace",
                                          "lookalike_match": False,
                                          "original_reference": "generic/Genres_Master.csv",
                                          "reference_field": "0",
                                          "skip_blank": False,
                                          "string_distance_threshold": 0.7
                                      },
                                      "rule_type": "Lookup",
                                      "field": "9"
                                  },
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": True,
                                          "default_value": "BaNonea",
                                          "fallback_mode": "do_not_replace",
                                          "lookalike_match": False,
                                          "original_reference": "generic/Genres_Master.csv",

                                          "reference_field": "0",
                                          "skip_blank": False,
                                          "string_distance_threshold": 0.7
                                      },
                                      "rule_type": "Lookup",
                                      "field": "10"
                                  },
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": True,
                                          "default_value": "BaNonea",
                                          "fallback_mode": "do_not_replace",
                                          "lookalike_match": False,
                                          "original_reference": "generic/Genres_Master.csv",

                                          "reference_field": "0",
                                          "skip_blank": False,
                                          "string_distance_threshold": 0.7
                                      },
                                      "rule_type": "Lookup",
                                      "field": "11"
                                  },
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": True,
                                          "default_value": "BaNonea",
                                          "fallback_mode": "do_not_replace",
                                          "lookalike_match": False,
                                          "original_reference": "generic/Genres_Master.csv",

                                          "reference_field": "0",
                                          "skip_blank": False,
                                          "string_distance_threshold": 0.7
                                      },
                                      "rule_type": "Lookup",
                                      "field": "12"
                                  },
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": True,
                                          "default_value": "BaNonea",
                                          "fallback_mode": "do_not_replace",
                                          "lookalike_match": False,
                                          "original_reference": "generic/Genres_Master.csv",

                                          "reference_field": "0",
                                          "skip_blank": False,
                                          "string_distance_threshold": 0.7
                                      },
                                      "rule_type": "Lookup",
                                      "field": "13"
                                  },
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": True,
                                          "default_value": "BaNonea",
                                          "fallback_mode": "do_not_replace",
                                          "lookalike_match": False,
                                          "original_reference": "generic/Genres_Master.csv",

                                          "reference_field": "0",
                                          "skip_blank": False,
                                          "string_distance_threshold": 0.7
                                      },
                                      "rule_type": "Lookup",
                                      "field": "14"
                                  }
                              ]
                          },
                          report=[
                              {

                                  'field': 'GroupTitle',
                                  'rule': 'Lookup',
                                  'number_records': 16,
                                  'category': 'replaced',
                                  'description': 'Best match',
                                  'df': [['Escape from L.A.', 'Escape from LA'],
                                         ['Executioners From Shaolin', 'Executed'],
                                         ['Five Shaolin Masters', 'Five Fingers'],
                                         ['Hungry for Change', 'Chor Aur Chand'],
                                         ['Invincible Shaolin', 'In Her Shoes'],
                                         ['Killer Legends', 'Beast Legends'],
                                         ['Knockout', 'AM Workout'],
                                         ['Nick of Time', 'Dick Tiger'],
                                         ['Pulp Fiction', 'Am Politician'],
                                         ['Reservoir Dogs', 'Firehouse Dog'],
                                         ['Soaked in Bleach', 'A Second Chance'],
                                         ['Survivor', 'Aurora'],
                                         ['The Truman Show', 'Fred The Show'],
                                         ['The Warriors', 'Fish Warrior'],
                                         ['Upstream Color', "Ceart's Coir"],
                                         ['Urban Cowboy', 'Carry On Cowboy']],

                              },
                              {

                                  'field': 'GroupTitle',
                                  'rule': 'Lookup',
                                  'number_records': 4,
                                  'category': 'ignored',
                                  'description': 'No changes made',
                                  'df': [['Martial Arts of Shaolin', 'Martial Arts of Shaolin'],
                                         ['Return To The 36th Chamber', 'Return To The 36th Chamber'],
                                         ['The Avenging Eagle', 'The Avenging Eagle'],
                                         ['The Five Venoms', 'The Five Venoms']],

                              },
                              {

                                  'field': 'Title',
                                  'rule': 'Lookup',
                                  'number_records': 16,
                                  'category': 'replaced',
                                  'description': 'Best match',
                                  'df': [['Escape from L.A.', 'Escape from LA'],
                                         ['Executioners From Shaolin', 'Executed'],
                                         ['Five Shaolin Masters', 'Five Fingers'],
                                         ['Hungry for Change', 'Chor Aur Chand'],
                                         ['Invincible Shaolin', 'In Her Shoes'],
                                         ['Killer Legends', 'Beast Legends'],
                                         ['Knockout', 'AM Workout'],
                                         ['Nick of Time', 'Dick Tiger'],
                                         ['Pulp Fiction', 'Am Politician'],
                                         ['Reservoir Dogs', 'Firehouse Dog'],
                                         ['Soaked in Bleach', 'A Second Chance'],
                                         ['Survivor', 'Aurora'],
                                         ['The Truman Show', 'Fred The Show'],
                                         ['The Warriors', 'Fish Warrior'],
                                         ['Upstream Color', "Ceart's Coir"],
                                         ['Urban Cowboy', 'Carry On Cowboy']],

                              },
                              {

                                  'field': 'Title',
                                  'rule': 'Lookup',
                                  'number_records': 4,
                                  'category': 'ignored',
                                  'description': 'No changes made',
                                  'df': [['Martial Arts of Shaolin', 'Martial Arts of Shaolin'],
                                         ['Return To The 36th Chamber', 'Return To The 36th Chamber'],
                                         ['The Avenging Eagle', 'The Avenging Eagle'],
                                         ['The Five Venoms', 'The Five Venoms']],

                              },
                              {

                                  'field': 'Genre2',
                                  'rule': 'Lookup',
                                  'number_records': 24,
                                  'category': 'replaced',
                                  'description': 'Best match',
                                  'df': [[' Adventures', 'Adventures'],
                                         [' Action Thrillers', 'Action Thrillers'],
                                         [' Action Thrillers', 'Action Thrillers'],
                                         [' Supernatural Horror Films', 'Supernatural Horror Films'],
                                         [' International Action & Adventure',
                                          'International Action & Adventure'],
                                         [' International Action & Adventure',
                                          'International Action & Adventure'],
                                         [' Social & Cultural Documentaries',
                                          'Social & Cultural Documentaries'],
                                         [' Documentaries', 'Documentaries'],
                                         [' International Action & Adventure',
                                          'International Action & Adventure'],
                                         [' Crime Documentaries', 'Crime Documentaries'],
                                         [' Films for ages 8 to 10', 'Films for ages 8 to 10'],
                                         [' International Action & Adventure',
                                          'International Action & Adventure'],
                                         [' Crime Thrillers', 'Crime Thrillers'],
                                         [' Crime Dramas', 'Crime Dramas'],
                                         [' Thrillers', 'Thrillers'],
                                         [' International Action & Adventure',
                                          'International Action & Adventure'],
                                         [' Crime Documentaries', 'Crime Documentaries'],
                                         [' Action Thrillers', 'Action Thrillers'],
                                         [' International Action & Adventure',
                                          'International Action & Adventure'],
                                         [' International Action & Adventure',
                                          'International Action & Adventure'],
                                         [' Dark Comedies', 'Dark Comedies'],
                                         [' Classic Action & Adventure', 'Classic Action & Adventure'],
                                         [' Independent Dramas', 'Independent Dramas'],
                                         [' Romantic Dramas', 'Romantic Dramas']],

                              },
                              {

                                  'field': 'Genre3',
                                  'rule': 'Lookup',
                                  'number_records': 21,
                                  'category': 'replaced',
                                  'description': 'Best match',
                                  'df': [[' Thrillers', 'Thrillers'],
                                         [' Sci-Fi Thrillers', 'Sci-Fi Thrillers'],
                                         [' Cult Films', 'Cult Films'],
                                         [' Sci-Fi & Fantasy', 'Sci-Fi & Fantasy'],
                                         [' Asian Action Films', 'Asian Action Films'],
                                         [' Asian Action Films', 'Asian Action Films'],
                                         [' Social & Cultural Documentaries',
                                          'Social & Cultural Documentaries'],
                                         [' Asian Action Films', 'Asian Action Films'],
                                         [' Social & Cultural Documentaries',
                                          'Social & Cultural Documentaries'],
                                         [' Films for ages 11 to 12', 'Films for ages 11 to 12'],
                                         [' Asian Action Films', 'Asian Action Films'],
                                         [' Independent Dramas', 'Independent Dramas'],
                                         [' Crime Thrillers', 'Crime Thrillers'],
                                         [' Asian Action Films', 'Asian Action Films'],
                                         [' Spy Action & Adventure', 'Spy Action & Adventure'],
                                         [' Asian Action Films', 'Asian Action Films'],
                                         [' Asian Action Films', 'Asian Action Films'],
                                         [' Dramas', 'Dramas'],
                                         [' Action Thrillers', 'Action Thrillers'],
                                         [' Independent Films', 'Independent Films'],
                                         [' Romantic Films', 'Romantic Films']],

                              },
                              {

                                  'field': 'Genre5',
                                  'rule': 'Lookup',
                                  'number_records': 2,
                                  'category': 'ignored',
                                  'description': 'No changes made',
                                  'df': [[None, None],
                                         [None, None]],

                              },
                              {

                                  'field': 'Genre6',
                                  'rule': 'Lookup',
                                  'number_records': 4,
                                  'category': 'ignored',
                                  'description': 'No changes made',
                                  'df': [[None, None],
                                         [None, None],
                                         [None, None],
                                         [None, None]],

                              },
                              {

                                  'field': 'Genre7',
                                  'rule': 'Lookup',
                                  'number_records': 11,
                                  'category': 'ignored',
                                  'description': 'No changes made',
                                  'df': [[None, None],
                                         [None, None],
                                         [None, None],
                                         [None, None],
                                         [None, None],
                                         [None, None],
                                         [None, None],
                                         [None, None],
                                         [None, None],
                                         [None, None],
                                         [None, None]],

                              },
                              {

                                  'field': 'Genre8',
                                  'rule': 'Lookup',
                                  'number_records': 20,
                                  'category': 'ignored',
                                  'description': 'No changes made',
                                  'df': [[None, None],
                                         [None, None],
                                         [None, None],
                                         [None, None],
                                         [None, None],
                                         [None, None],
                                         [None, None],
                                         [None, None],
                                         [None, None],
                                         [None, None],
                                         [None, None],
                                         [None, None],
                                         [None, None],
                                         [None, None],
                                         [None, None],
                                         [None, None],
                                         [None, None],
                                         [None, None],
                                         [None, None],
                                         [None, None]],

                              },
                          ])



# (c) 2012-2018 Dativa, all rights reserved
# -----------------------------------------
# Usage subject to license agreement
# hello@dativa.com for more information

from dativa.tools.pandas import CSVHandler
from dativa.scrubber.tests import _BaseTest
from dativa.scrubber import Scrubber, ScrubberValidationError


class AnonymizationTests(_BaseTest):

    def test_errors_hash(self):
        self._test_filter(dirty_file="anonymization/emails.csv",
                          clean_file="anonymization/emails.csv",
                          config={
                              "rules": [{
                                  "rule_type": "String",
                                  "field": "from",
                                  "params": {
                                      "minimum_length": 5,
                                      "maximum_length": 1024,
                                      "regex": "[^@]+@[^\\.]..*[^\\.]",
                                      "fallback_mode": "remove_record",
                                      "hash": True,
                                      "salt": "A9788@%y%m%d",
                                      "salt_date_field": "date"
                                  }}]
                          },
                          expected_error=ScrubberValidationError)

        self._test_filter(dirty_file="anonymization/emails.csv",
                          clean_file="anonymization/emails.csv",
                          config={
                              "rules": [{
                                  "rule_type": "String",
                                  "field": "from",
                                  "params": {
                                      "minimum_length": 5,
                                      "maximum_length": 1024,
                                      "regex": "[^@]+@[^\\.]..*[^\\.]",
                                      "fallback_mode": "remove_record",
                                      "hash": True,
                                      "salt": "A9788@",
                                      "salt_date_field": "date",
                                      "salt_date_format": "%Y-%m-%d %H:%M:%S"
                                  }}]
                          },
                          expected_error=ScrubberValidationError)

        self._test_filter(dirty_file="anonymization/emails.csv",
                          clean_file="anonymization/emails.csv",
                          config={
                              "rules": [{
                                  "rule_type": "String",
                                  "field": "from",
                                  "params": {
                                      "minimum_length": 5,
                                      "maximum_length": 1024,
                                      "regex": "[^@]+@[^\\.]..*[^\\.]",
                                      "fallback_mode": "remove_record",
                                      "hash": True,
                                      "salt": "1234567891%Y%m%d",
                                      "salt_date_field": "date",
                                      "salt_date_format": "%Y-%m-%d %H:%M:%S"
                                  }}]
                          },
                          expected_error=ScrubberValidationError)

        self._test_filter(dirty_file="anonymization/emails.csv",
                          clean_file="anonymization/emails.csv",
                          config={
                              "rules": [{
                                  "rule_type": "String",
                                  "field": "from",
                                  "params": {
                                      "minimum_length": 5,
                                      "maximum_length": 1024,
                                      "regex": "[^@]+@[^\\.]..*[^\\.]",
                                      "fallback_mode": "remove_record",
                                      "hash": True,
                                      "salt": "A9788@%y%m%d",
                                      "salt_date_field": "data",
                                      "salt_date_format": "%Y-%m-%d %H:%M:%S"
                                  }}]
                          },
                          expected_error=ScrubberValidationError)

    def test_errors_encrypt(self):
        self._test_filter(dirty_file="anonymization/emails.csv",
                          clean_file="anonymization/emails.csv",
                          config={
                              "rules": [{
                                  "rule_type": "String",
                                  "field": "from",
                                  "params": {
                                      "minimum_length": 5,
                                      "maximum_length": 1024,
                                      "regex": "[^@]+@[^\\.]..*[^\\.]",
                                      "fallback_mode": "remove_record",
                                      "encrypt": True
                                  }}]
                          },
                          expected_error=ScrubberValidationError)

        self._test_filter(dirty_file="anonymization/emails.csv",
                          clean_file="anonymization/emails.csv",
                          config={
                              "rules": [{
                                  "rule_type": "String",
                                  "field": "from",
                                  "params": {
                                      "minimum_length": 5,
                                      "maximum_length": 1024,
                                      "regex": "[^@]+@[^\\.]..*[^\\.]",
                                      "fallback_mode": "remove_record",
                                      "encrypt": True,
                                      "public_key": "not a public key"
                                  }}]
                          },
                          expected_error=ScrubberValidationError)

        self._test_filter(dirty_file="anonymization/emails.csv",
                          clean_file="anonymization/emails.csv",
                          config={
                              "rules": [{
                                  "rule_type": "String",
                                  "field": "from",
                                  "params": {
                                      "minimum_length": 5,
                                      "maximum_length": 1024,
                                      "regex": "[^@]+@[^\\.]..*[^\\.]",
                                      "fallback_mode": "remove_record",
                                      "encrypt": True,
                                      "public_key": """-----BEGIN PUBLIC KEY-----
                                                       MIGfMA0GCSqGSIb3DQEBAQUAA4GNADCBiQKBgQDDmPPm5UC8rXn4uX37m4tN/j4T
                                                       MAhUVyxN7V7QxMF3HDg5rkl/Ju53DPJbv59TCvlTCXw1ihp9asVyyYpCqrsKCh10
                                                       sZI0kIrkizlKaB/20Q4P1kYOCgv4Cwds7Iu2y0TFwDosK9a7MPR9IksL7QRWKjD0
                                                       DoNemKEpyCt2dZTaQwIDAQAB
                                                       -----END PUBLIC KEY-----""",
                                      "random_string": "1234567891%Y%m%d",
                                      "random_string_date_field": "date"
                                  }}]
                          },
                          expected_error=ScrubberValidationError)

        self._test_filter(dirty_file="anonymization/emails.csv",
                          clean_file="anonymization/emails.csv",
                          config={
                              "rules": [{
                                  "rule_type": "String",
                                  "field": "from",
                                  "params": {
                                      "minimum_length": 5,
                                      "maximum_length": 1024,
                                      "regex": "[^@]+@[^\\.]..*[^\\.]",
                                      "fallback_mode": "remove_record",
                                      "encrypt": True,
                                      "public_key": """-----BEGIN PUBLIC KEY-----
                                                       MIGfMA0GCSqGSIb3DQEBAQUAA4GNADCBiQKBgQDDmPPm5UC8rXn4uX37m4tN/j4T
                                                       MAhUVyxN7V7QxMF3HDg5rkl/Ju53DPJbv59TCvlTCXw1ihp9asVyyYpCqrsKCh10
                                                       sZI0kIrkizlKaB/20Q4P1kYOCgv4Cwds7Iu2y0TFwDosK9a7MPR9IksL7QRWKjD0
                                                       DoNemKEpyCt2dZTaQwIDAQAB
                                                       -----END PUBLIC KEY-----""",
                                      "random_string_date_field": "date",
                                      "random_string_date_format": "%Y-%m-%d %H:%M:%S",
                                      "random_string": "1234567891"
                                  }}]
                          },
                          expected_error=ScrubberValidationError)

        self._test_filter(dirty_file="anonymization/emails.csv",
                          clean_file="anonymization/emails.csv",
                          config={
                              "rules": [{
                                  "rule_type": "String",
                                  "field": "from",
                                  "params": {
                                      "minimum_length": 5,
                                      "maximum_length": 1024,
                                      "regex": "[^@]+@[^\\.]..*[^\\.]",
                                      "fallback_mode": "remove_record",
                                      "encrypt": True,
                                      "random_string": "A" * 18 + "%Y",
                                      "random_string_date_field": "date",
                                      "random_string_date_format": "%Y-%m-%d %H:%M:%S",
                                      "public_key": """-----BEGIN PUBLIC KEY-----
                                                       MIGfMA0GCSqGSIb3DQEBAQUAA4GNADCBiQKBgQDDmPPm5UC8rXn4uX37m4tN/j4T
                                                       MAhUVyxN7V7QxMF3HDg5rkl/Ju53DPJbv59TCvlTCXw1ihp9asVyyYpCqrsKCh10
                                                       sZI0kIrkizlKaB/20Q4P1kYOCgv4Cwds7Iu2y0TFwDosK9a7MPR9IksL7QRWKjD0
                                                       DoNemKEpyCt2dZTaQwIDAQAB
                                                       -----END PUBLIC KEY-----"""
                                  }}]
                          },
                          expected_error=ScrubberValidationError)

        self._test_filter(dirty_file="anonymization/emails.csv",
                          clean_file="anonymization/emails.csv",
                          config={
                              "rules": [{
                                  "rule_type": "String",
                                  "field": "from",
                                  "params": {
                                      "minimum_length": 5,
                                      "maximum_length": 1024,
                                      "regex": "[^@]+@[^\\.]..*[^\\.]",
                                      "fallback_mode": "remove_record",
                                      "encrypt": True,
                                      "public_key": """-----BEGIN PUBLIC KEY-----
                                                       MIGfMA0GCSqGSIb3DQEBAQUAA4GNADCBiQKBgQDDmPPm5UC8rXn4uX37m4tN/j4T
                                                       MAhUVyxN7V7QxMF3HDg5rkl/Ju53DPJbv59TCvlTCXw1ihp9asVyyYpCqrsKCh10
                                                       sZI0kIrkizlKaB/20Q4P1kYOCgv4Cwds7Iu2y0TFwDosK9a7MPR9IksL7QRWKjD0
                                                       DoNemKEpyCt2dZTaQwIDAQAB
                                                       -----END PUBLIC KEY-----""",
                                      "random_string_date_field": "date_not_there",
                                      "random_string_date_format": "%Y-%m-%d %H:%M:%S",
                                      "random_string": "1234567891%y%m%d"
                                  }}]
                          },
                          expected_error=ScrubberValidationError)

    def test_errors_decrypt(self):
        self._test_filter(dirty_file="anonymization/emails.csv",
                          clean_file="anonymization/emails.csv",
                          config={
                              "rules": [{
                                  "rule_type": "String",
                                  "field": "from",
                                  "params": {
                                      "minimum_length": 5,
                                      "maximum_length": 1024,
                                      "regex": "[^@]+@[^\\.]..*[^\\.]",
                                      "fallback_mode": "remove_record",
                                      "decrypt": True
                                  }}]
                          },
                          expected_error=ScrubberValidationError)

        self._test_filter(dirty_file="anonymization/emails.csv",
                          clean_file="anonymization/emails.csv",
                          config={
                              "rules": [{
                                  "rule_type": "String",
                                  "field": "from",
                                  "params": {
                                      "minimum_length": 5,
                                      "maximum_length": 1024,
                                      "regex": "[^@]+@[^\\.]..*[^\\.]",
                                      "fallback_mode": "remove_record",
                                      "decrypt": True,
                                      "private_key": "not a private key"
                                  }}]
                          },
                          expected_error=ScrubberValidationError)

    def test_email_base(self):
        self._test_filter(dirty_file="anonymization/emails.csv",
                          clean_file="anonymization/emails.csv",
                          config={
                              "rules": [{
                                  "rule_type": "String",
                                  "field": "to",
                                  "params": {
                                      "minimum_length": 5,
                                      "maximum_length": 1024,
                                      "regex": "[^@]+@[^\\.]..*[^\\.]",
                                      "fallback_mode": "remove_record"
                                  }},
                                  {
                                      "rule_type": "String",
                                      "field": "from",
                                      "params": {
                                          "minimum_length": 5,
                                          "maximum_length": 1024,
                                          "regex": "[^@]+@[^\\.]..*[^\\.]",
                                          "fallback_mode": "remove_record"
                                      }},
                                  {
                                      "rule_type": "Date",
                                      "field": "date",
                                      "params": {
                                          "date_format": "%Y-%m-%d %H:%M:%S",
                                          "fallback_mode": "remove_record",
                                          "range_minimum": "1970-01-01 00:00:00",
                                          "range_maximum": "2100-01-01 00:00:00",
                                      }},
                                  {
                                      "rule_type": "String",
                                      "field": "subject",
                                      "params": {
                                          "minimum_length": 0,
                                          "maximum_length": 1024,
                                          "fallback_mode": "remove_record"
                                      }}]
                          },
                          report=[])

    def test_email_hash(self):
        self._test_filter(dirty_file="anonymization/emails.csv",
                          clean_file="anonymization/emails_hash.csv",
                          config={
                              "rules": [{
                                  "rule_type": "String",
                                  "field": "to",
                                  "params": {
                                      "minimum_length": 5,
                                      "maximum_length": 1024,
                                      "regex": "[^@]+@[^\\.]..*[^\\.]",
                                      "fallback_mode": "remove_record",
                                      "hash": True,
                                      "salt": "A97ASFNKJ#&"
                                  }},
                                  {
                                      "rule_type": "String",
                                      "field": "from",
                                      "params": {
                                          "minimum_length": 5,
                                          "maximum_length": 1024,
                                          "regex": "[^@]+@[^\\.]..*[^\\.]",
                                          "fallback_mode": "remove_record",
                                          "hash": True,
                                          "salt": "A97ASFNKJ#&"
                                      }},
                                  {
                                      "rule_type": "Date",
                                      "field": "date",
                                      "params": {
                                          "date_format": "%Y-%m-%d %H:%M:%S",
                                          "fallback_mode": "remove_record",
                                          "range_minimum": "1970-01-01 00:00:00",
                                          "range_maximum": "2100-01-01 00:00:00",
                                      }},
                                  {
                                      "rule_type": "String",
                                      "field": "subject",
                                      "params": {
                                          "minimum_length": 0,
                                          "maximum_length": 1024,
                                          "fallback_mode": "remove_record"
                                      }}]
                          },
                          report=[
                              {

                                  'field': 'to_clean',
                                  'rule': 'String',
                                  'number_records': 6,
                                  'category': 'modified',
                                  'description': 'Hashed all values',
                                  'df': [['3088b61a1fb2b613'],
                                         ['754d2dc7563c36f1'],
                                         ['4b6c888fa57e29dd'],
                                         ['3088b61a1fb2b613'],
                                         ['754d2dc7563c36f1'],
                                         ['4b6c888fa57e29dd']],

                              },
                              {

                                  'field': 'from_clean',
                                  'rule': 'String',
                                  'number_records': 6,
                                  'category': 'modified',
                                  'description': 'Hashed all values',
                                  'df': [['754d2dc7563c36f1'],
                                         ['3088b61a1fb2b613'],
                                         ['3088b61a1fb2b613'],
                                         ['754d2dc7563c36f1'],
                                         ['3088b61a1fb2b613'],
                                         ['754d2dc7563c36f1']],

                              },
                          ])

    def test_email_hash_rotating(self):
        self._test_filter(dirty_file="anonymization/emails.csv",
                          clean_file="anonymization/emails_rotating_hash.csv",
                          config={
                              "rules": [{
                                  "rule_type": "String",
                                  "field": "to",
                                  "params": {
                                      "minimum_length": 5,
                                      "maximum_length": 1024,
                                      "regex": "[^@]+@[^\\.]..*[^\\.]",
                                      "fallback_mode": "remove_record",
                                      "hash": True,
                                      "salt": "A9788@%y%m%d",
                                      "salt_date_field": "date",
                                      "salt_date_format": "%Y-%m-%d %H:%M:%S"
                                  }},
                                  {
                                      "rule_type": "String",
                                      "field": "from",
                                      "params": {
                                          "minimum_length": 5,
                                          "maximum_length": 1024,
                                          "regex": "[^@]+@[^\\.]..*[^\\.]",
                                          "fallback_mode": "remove_record",
                                          "hash": True,
                                          "salt": "A9788@%y%m%d",
                                          "salt_date_field": "date",
                                          "salt_date_format": "%Y-%m-%d %H:%M:%S"
                                      }},
                                  {
                                      "rule_type": "Date",
                                      "field": "date",
                                      "params": {
                                          "date_format": "%Y-%m-%d %H:%M:%S",
                                          "fallback_mode": "remove_record",
                                          "range_minimum": "1970-01-01 00:00:00",
                                          "range_maximum": "2100-01-01 00:00:00",
                                      }},
                                  {
                                      "rule_type": "String",
                                      "field": "subject",
                                      "params": {
                                          "minimum_length": 0,
                                          "maximum_length": 1024,
                                          "fallback_mode": "remove_record"
                                      }}]
                          },
                          report=[
                              {

                                  'field': 'to_clean',
                                  'rule': 'String',
                                  'number_records': 6,
                                  'category': 'modified',
                                  'description': 'Hashed all values to salt date',
                                  'df': [['1990-10-12 18:32:42', 'dc7a794b4d88ed33'],
                                         ['1990-10-12 18:35:42', '862fc7f148bdc155'],
                                         ['1990-10-12 18:37:22', '48d01575e5e1c26d'],
                                         ['1990-10-12 18:45:42', 'dc7a794b4d88ed33'],
                                         ['1990-10-13 08:23:11', '9b64a5446478edea'],
                                         ['1990-10-13 11:22:33', 'd13fdff7bc26425c']],

                              },
                              {

                                  'field': 'from_clean',
                                  'rule': 'String',
                                  'number_records': 6,
                                  'category': 'modified',
                                  'description': 'Hashed all values to salt date',
                                  'df': [['1990-10-12 18:32:42', '862fc7f148bdc155'],
                                         ['1990-10-12 18:35:42', 'dc7a794b4d88ed33'],
                                         ['1990-10-12 18:37:22', 'dc7a794b4d88ed33'],
                                         ['1990-10-12 18:45:42', '862fc7f148bdc155'],
                                         ['1990-10-13 08:23:11', 'e640150619112ae9'],
                                         ['1990-10-13 11:22:33', '9b64a5446478edea']],

                              },
                          ])

    def test_email_encryption(self):
        df_1 = [[
            'upbFgaZ1gIbH8ph7uY8LFTCHG/yzPBxKwIguiHW/0TwK3I46QLZJwWj/YdB9kQ04RDuSUEm5+'
            'J5sti4hwZbSFU4WUAylZn82kw+dGSSUqPpSjn+MwQE58zTUUVULzBX0oceY5psr2Aur3EbzTulLxWKkkx73HcsFHDhKD6qdx84='],
            [
                'OGoDXfksvrNzCz09u1dGZcYY1gNWLMIZAdb32l+mJPrY7KeXDQkJ7NnE42Q2AtbV0r76yrHThP7E2Wj'
                '9hUc8Fgx3cgOPMxHnXnDNFcZL1wrk0WmpIqaDUtFG8a2/fzpfi4RWG9l/F6jUn7Kat4WQpa0UHVgxL21gTTGAnStZqyU='],
            [
                'SukuZeemYwKNV1trdUbAoGtODop3RvU4EKNXUVR/oonm7uu+fiqdSsPYrXiVHztfXQbZrFz69Gt0LfHwU7VqmSg4uMg'
                'K8/sqQaHUSDUxPtY2AEB7M1EsqrdsnzvDZzf87E/R7JC+ZYW7kAZXpjHWE5LA0nZH/j6o4r+jlRFvxX8='],
            [
                'upbFgaZ1gIbH8ph7uY8LFTCHG/yzPBxKwIguiHW/0TwK3I46QLZJwWj/YdB9kQ04RDuSUEm5+J5sti4hwZbSFU4WUAyl'
                'Zn82kw+dGSSUqPpSjn+MwQE58zTUUVULzBX0oceY5psr2Aur3EbzTulLxWKkkx73HcsFHDhKD6qdx84='],
            [
                'OGoDXfksvrNzCz09u1dGZcYY1gNWLMIZAdb32l+mJPrY7KeXDQkJ7NnE42Q2AtbV0r76yrHThP7E2Wj9hUc8'
                'Fgx3cgOPMxHnXnDNFcZL1wrk0WmpIqaDUtFG8a2/fzpfi4RWG9l/F6jUn7Kat4WQpa0UHVgxL21gTTGAnStZqyU='],
            [
                'SukuZeemYwKNV1trdUbAoGtODop3RvU4EKNXUVR/oonm7uu+fiqdSsPYrXiVHztfXQbZrFz69Gt0LfHwU7VqmSg4uMgK8'
                '/sqQaHUSDUxPtY2AEB7M1EsqrdsnzvDZzf87E/R7JC+ZYW7kAZXpjHWE5LA0nZH/j6o4r+jlRFvxX8=']]

        df_2 = [[
            'OGoDXfksvrNzCz09u1dGZcYY1gNWLMIZAdb32l+mJPrY7KeXDQkJ7NnE42Q2AtbV0r76yrHThP7E2Wj9hUc8F'
            'gx3cgOPMxHnXnDNFcZL1wrk0WmpIqaDUtFG8a2/fzpfi4RWG9l'
            '/F6jUn7Kat4WQpa0UHVgxL21gTTGAnStZqyU='],
            [
                'upbFgaZ1gIbH8ph7uY8LFTCHG/yzPBxKwIguiHW/0TwK3I46QLZJwWj/YdB9kQ04RDuSUEm5+J5sti4hwZbSFU4WUAylZn82kw+'
                'dGSSUqPpSjn+MwQE58zTUUVULzBX0oceY5psr2Aur3EbzTulLxWKkkx73HcsFHDhKD6qdx84='],
            [
                'upbFgaZ1gIbH8ph7uY8LFTCHG/yzPBxKwIguiHW/0TwK3I46QLZJwWj/YdB9kQ04RDuSUEm5+J5sti4hwZbSFU4WUAylZn82kw+d'
                'GSSUqPpSjn+MwQE58zTUUVULzBX0oceY5psr2Aur3EbzTulLxWKkkx73HcsFHDhKD6qdx84='],
            [
                'OGoDXfksvrNzCz09u1dGZcYY1gNWLMIZAdb32l+mJPrY7KeXDQkJ7NnE42Q2AtbV0r76yrHThP7E2Wj9hUc8F'
                'gx3cgOPMxHnXnDNFcZL1wrk0WmpIqaDUtFG8a2/fzpfi4RWG9l/F6jUn7Kat4WQpa0UHVgxL21gTTGAnStZqyU='],
            [
                'upbFgaZ1gIbH8ph7uY8LFTCHG/yzPBxKwIguiHW/0TwK3I46QLZJwWj/YdB9kQ04RDuSUEm5+J5sti4hwZbSFU4WUAylZn82kw+'
                'dGSSUqPpSjn+MwQE58zTUUVULzBX0oceY5psr2Aur3EbzTulLxWKkkx73HcsFHDhKD6qdx84='],
            [
                'OGoDXfksvrNzCz09u1dGZcYY1gNWLMIZAdb32l+mJPrY7KeXDQkJ7NnE42Q2AtbV0r76yrHThP7E2Wj9hUc8F'
                'gx3cgOPMxHnXnDNFcZL1wrk0WmpIqaDUtFG8a2/fzpfi4RWG9l/F6jUn7Kat4WQpa0UHVgxL21gTTGAnStZqyU=']]

        self._test_filter(dirty_file="anonymization/emails.csv",
                          clean_file="anonymization/emails_encrypted.csv",
                          config={
                              "rules": [{
                                  "rule_type": "String",
                                  "field": "to",
                                  "params": {
                                      "minimum_length": 5,
                                      "maximum_length": 1024,
                                      "regex": "[^@]+@[^\\.]..*[^\\.]",
                                      "fallback_mode": "remove_record",
                                      "encrypt": True,
                                      "random_string": "",
                                      "public_key": """-----BEGIN PUBLIC KEY-----
                               MIGfMA0GCSqGSIb3DQEBAQUAA4GNADCBiQKBgQDDmPPm5UC8rXn4uX37m4tN/j4T
                               MAhUVyxN7V7QxMF3HDg5rkl/Ju53DPJbv59TCvlTCXw1ihp9asVyyYpCqrsKCh10
                               sZI0kIrkizlKaB/20Q4P1kYOCgv4Cwds7Iu2y0TFwDosK9a7MPR9IksL7QRWKjD0
                               DoNemKEpyCt2dZTaQwIDAQAB
                               -----END PUBLIC KEY-----"""
                                  }},
                                  {
                                      "rule_type": "String",
                                      "field": "from",
                                      "params": {
                                          "minimum_length": 5,
                                          "maximum_length": 1024,
                                          "regex": "[^@]+@[^\\.]..*[^\\.]",
                                          "fallback_mode": "remove_record",
                                          "encrypt": True,
                                          "random_string": "",
                                          "public_key": """-----BEGIN PUBLIC KEY-----
                               MIGfMA0GCSqGSIb3DQEBAQUAA4GNADCBiQKBgQDDmPPm5UC8rXn4uX37m4tN/j4T
                               MAhUVyxN7V7QxMF3HDg5rkl/Ju53DPJbv59TCvlTCXw1ihp9asVyyYpCqrsKCh10
                               sZI0kIrkizlKaB/20Q4P1kYOCgv4Cwds7Iu2y0TFwDosK9a7MPR9IksL7QRWKjD0
                               DoNemKEpyCt2dZTaQwIDAQAB
                               -----END PUBLIC KEY-----"""
                                      }},
                                  {
                                      "rule_type": "Date",
                                      "field": "date",
                                      "params": {
                                          "date_format": "%Y-%m-%d %H:%M:%S",
                                          "fallback_mode": "remove_record",
                                          "range_minimum": "1970-01-01 00:00:00",
                                          "range_maximum": "2100-01-01 00:00:00",
                                      }},
                                  {
                                      "rule_type": "String",
                                      "field": "subject",
                                      "params": {
                                          "minimum_length": 0,
                                          "maximum_length": 1024,
                                          "fallback_mode": "remove_record"
                                      }}]
                          },
                          report=[
                              {

                                  'field': 'to_clean',
                                  'rule': 'String',
                                  'number_records': 6,
                                  'category': 'modified',
                                  'description': 'Encrypted values',
                                  'df': df_1,

                              },
                              {

                                  'field': 'from_clean',
                                  'rule': 'String',
                                  'number_records': 6,
                                  'category': 'modified',
                                  'description': 'Encrypted values',
                                  'df': df_2,

                              },
                          ])

    def test_email_encryption_rotating(self):
        df_1 = [[
            'm8M0AWJgi3a0ntCDQMTCoMrAa9WbM9qW1soucDTMzXEa/gnHLSiEaTZ+x1mi4bag7JaRNOYJsXBQDTOwEHh+CjvkF/nNK'
            '+y7ViboXCWHhi9UpF7HePBCFpNtHATguIQQrQzvHq3wYPXax58u6tQrxrCXVDfBt7+pCpUODwz6KP4=',
            '1990-10-12 18:32:42'],
            [
                'Y0pS5yTkR7iXTnp6ZbnRxTI0aEWPJoFjb8we6aasooXoK3/GOsmmaGt6pBnN6GUFUCMly+iOMjX5veW340Xp'
                '+f6o783RPCMiaroYJPMgenu4w3b8SZWP4YH0Nled3Lmn6a/iBuwDNyUydQ0dmKCRxoU8YGqQ8bE8MXf+Ot38P7w=',
                '1990-10-12 18:35:42'],
            [
                'iBlLHVl9PlncLgbtbRAZJix/q005QWlOTBhrYVZhBtG5wRy1gwX1DqZ2bSf1pxFhBggrxAlwKyOFO04t3qBrtqZwpKRK1oCb'
                '+z45WDSc7Cdvi5MwgmH5EG34gOxcI3Ms36KsZ5m35VwXHU8B8t3JucipQcHlwA+CdzGdBmRHKeo=',
                '1990-10-12 18:37:22'],
            [
                'm8M0AWJgi3a0ntCDQMTCoMrAa9WbM9qW1soucDTMzXEa/gnHLSiEaTZ+x1mi4bag7JaRNOYJsXBQDTOwEHh+CjvkF/nNK'
                '+y7ViboXCWHhi9UpF7HePBCFpNtHATguIQQrQzvHq3wYPXax58u6tQrxrCXVDfBt7+pCpUODwz6KP4=',
                '1990-10-12 18:45:42'],
            [
                'If5mhgQMGFQCNxG5i5anZ0VndyfGAXHKgv3hBmmIbrYFvRB6USKGBeoOGnngRER1kydAdN52i4jJOG4Q6qKsuDAayn9M'
                '/DkNLLIsqJ0GXcndQYDz9e17ExsJJ4ZxVpVhvEZSSGLzIUNNH9dssNZ8Huz/ERFNGKA+ENdTsR/fm7A=',
                '1990-10-13 08:23:11'],
            [
                'YBtY8GvdTlUOxzdFxatzeHu8Ze3LPGcDueLqbPd6yK9iGibgjI/oTay/MQgPH8ybAwQ50Uv4/HxnjwTfPZTbzSXsJ'
                '+YPIjkZAFyDK4pjUepFKWdGESiQ/6OCt4MuPPN5dwk4KDHcdhrgabN7cesBP1X5+QveierexMjDtMGValo=',
                '1990-10-13 11:22:33']]

        df_2 = [[
            'Y0pS5yTkR7iXTnp6ZbnRxTI0aEWPJoFjb8we6aasooXoK3/GOsmmaGt6pBnN6GUFUCMly+iOMjX5veW340Xp'
            '+f6o783RPCMiaroYJPMgenu4w3b8SZWP4YH0Nled3Lmn6a/iBuwDNyUydQ0dmKCRxoU8YGqQ8bE8MXf+Ot38P7w=',
            '1990-10-12 18:32:42'],
            [
                'm8M0AWJgi3a0ntCDQMTCoMrAa9WbM9qW1soucDTMzXEa/gnHLSiEaTZ+x1mi4bag7JaRNOYJsXBQDTOwEHh+CjvkF/nNK'
                '+y7ViboXCWHhi9UpF7HePBCFpNtHATguIQQrQzvHq3wYPXax58u6tQrxrCXVDfBt7+pCpUODwz6KP4=',
                '1990-10-12 18:35:42'],
            [
                'm8M0AWJgi3a0ntCDQMTCoMrAa9WbM9qW1soucDTMzXEa/gnHLSiEaTZ+x1mi4bag7JaRNOYJsXBQDTOwEHh+CjvkF/nNK'
                '+y7ViboXCWHhi9UpF7HePBCFpNtHATguIQQrQzvHq3wYPXax58u6tQrxrCXVDfBt7+pCpUODwz6KP4=',
                '1990-10-12 18:37:22'],
            [
                'Y0pS5yTkR7iXTnp6ZbnRxTI0aEWPJoFjb8we6aasooXoK3/GOsmmaGt6pBnN6GUFUCMly+iOMjX5veW340Xp'
                '+f6o783RPCMiaroYJPMgenu4w3b8SZWP4YH0Nled3Lmn6a/iBuwDNyUydQ0dmKCRxoU8YGqQ8bE8MXf+Ot38P7w=',
                '1990-10-12 18:45:42'],
            [
                'Zypg4OEJFFik9fAB/SujORdjIagGlZcksn/+eTPeHm5IdTOIWm77aa/ij+3H5lqQEz1LoVWst8V5TtGJ/gMq63fAYsoj3pN'
                '+vNIM1lKpGrybO1JagDsPHP2o6GESjUrFKrfkH2OzdWDchHCdDksx8b/eMtOI8oTakncaZMwcPYA=',
                '1990-10-13 08:23:11'],
            [
                'If5mhgQMGFQCNxG5i5anZ0VndyfGAXHKgv3hBmmIbrYFvRB6USKGBeoOGnngRER1kydAdN52i4jJOG4Q6qKsuDAayn9M'
                '/DkNLLIsqJ0GXcndQYDz9e17ExsJJ4ZxVpVhvEZSSGLzIUNNH9dssNZ8Huz/ERFNGKA+ENdTsR/fm7A=',
                '1990-10-13 11:22:33']]
        self._test_filter(dirty_file="anonymization/emails.csv",
                          clean_file="anonymization/emails_rotating_encrypted.csv",
                          config={
                              "rules": [{
                                  "rule_type": "String",
                                  "field": "to",
                                  "params": {
                                      "minimum_length": 5,
                                      "maximum_length": 1024,
                                      "regex": "[^@]+@[^\\.]..*[^\\.]",
                                      "fallback_mode": "remove_record",
                                      "encrypt": True,
                                      "public_key": """-----BEGIN PUBLIC KEY-----
                               MIGfMA0GCSqGSIb3DQEBAQUAA4GNADCBiQKBgQDDmPPm5UC8rXn4uX37m4tN/j4T
                               MAhUVyxN7V7QxMF3HDg5rkl/Ju53DPJbv59TCvlTCXw1ihp9asVyyYpCqrsKCh10
                               sZI0kIrkizlKaB/20Q4P1kYOCgv4Cwds7Iu2y0TFwDosK9a7MPR9IksL7QRWKjD0
                               DoNemKEpyCt2dZTaQwIDAQAB
                               -----END PUBLIC KEY-----""",
                                      "random_string": "A9788@%y%m%d",
                                      "random_string_date_field": "date",
                                      "random_string_date_format": "%Y-%m-%d %H:%M:%S"
                                  }},
                                  {
                                      "rule_type": "String",
                                      "field": "from",
                                      "params": {
                                          "minimum_length": 5,
                                          "maximum_length": 1024,
                                          "regex": "[^@]+@[^\\.]..*[^\\.]",
                                          "fallback_mode": "remove_record",
                                          "encrypt": True,
                                          "public_key": """-----BEGIN PUBLIC KEY-----
                               MIGfMA0GCSqGSIb3DQEBAQUAA4GNADCBiQKBgQDDmPPm5UC8rXn4uX37m4tN/j4T
                               MAhUVyxN7V7QxMF3HDg5rkl/Ju53DPJbv59TCvlTCXw1ihp9asVyyYpCqrsKCh10
                               sZI0kIrkizlKaB/20Q4P1kYOCgv4Cwds7Iu2y0TFwDosK9a7MPR9IksL7QRWKjD0
                               DoNemKEpyCt2dZTaQwIDAQAB
                               -----END PUBLIC KEY-----""",
                                          "random_string": "A9788@%y%m%d",
                                          "random_string_date_field": "date",
                                          "random_string_date_format": "%Y-%m-%d %H:%M:%S"
                                      }},
                                  {
                                      "rule_type": "Date",
                                      "field": "date",
                                      "params": {
                                          "date_format": "%Y-%m-%d %H:%M:%S",
                                          "fallback_mode": "remove_record",
                                          "range_minimum": "1970-01-01 00:00:00",
                                          "range_maximum": "2100-01-01 00:00:00",
                                      }},
                                  {
                                      "rule_type": "String",
                                      "field": "subject",
                                      "params": {
                                          "minimum_length": 0,
                                          "maximum_length": 1024,
                                          "fallback_mode": "remove_record"
                                      }}]
                          },
                          report=[
                              {

                                  'field': 'to_clean',
                                  'rule': 'String',
                                  'number_records': 6,
                                  'category': 'modified',
                                  'description': 'Encrypted values to date field',
                                  'df': df_1,

                              },
                              {

                                  'field': 'from_clean',
                                  'rule': 'String',
                                  'number_records': 6,
                                  'category': 'modified',
                                  'description': 'Encrypted values to date field',
                                  'df': df_2,

                              },
                          ])

    def test_email_decryption(self):
        self._test_filter(dirty_file="anonymization/emails_rotating_encrypted.csv",
                          clean_file="anonymization/emails.csv",
                          config={
                              "rules": [{
                                  "rule_type": "String",
                                  "field": "to",
                                  "params": {
                                      "minimum_length": 5,
                                      "maximum_length": 1024,
                                      "regex": "[^@]+@[^\\.]..*[^\\.]",
                                      "fallback_mode": "remove_record",
                                      "decrypt": True,
                                      "private_key": """-----BEGIN RSA PRIVATE KEY-----
                                MIICXQIBAAKBgQDDmPPm5UC8rXn4uX37m4tN/j4TMAhUVyxN7V7QxMF3HDg5rkl/
                                Ju53DPJbv59TCvlTCXw1ihp9asVyyYpCqrsKCh10sZI0kIrkizlKaB/20Q4P1kYO
                                Cgv4Cwds7Iu2y0TFwDosK9a7MPR9IksL7QRWKjD0DoNemKEpyCt2dZTaQwIDAQAB
                                AoGBAIxvLv8irp5JN3+7Ppb+EMDIwCzqmbBkfmzc9uyRuA4a9suDNUXT3/ceFoyF
                                QpFPGb/i2YlfRLGzjVxjE5/WET9bf//ATOkZyn2sl4jBGs9WIF4by+246E+dfol9
                                /326rSpJjVeklmELU3nJp4ViZzYX+Cwc0rucpZLMqLUUFbvhAkEAzP+WSV2cp5w4
                                E1OBDM7cyq3lneWP0N1OdteZPZlJT0ojsQuoggqvcUMeC/eSoaHyZGN7GS4OVtaC
                                BR++kkPiCwJBAPRCnk4FmeLtnDpTX7KV8a766jSqkDzqEKVVqoGfLYCe0uaWFcwK
                                MKBik0GJ6tFLg9JvJMbP1+eD8+2uQD7rg6kCQDNghyDiBkX3oBIv5nL4UVu2k4qs
                                IwwcuvKL/Er05OurUCCqJFRbKzc+tAQZyzUZKm/AgvR/l3ZqEnIIT7HGs5sCQDej
                                NQvwmqzmEr/2XcYAAZ0p6k80ysYVStVePghoiaTSiJedeDmR2KGv0nsLP0GNQemd
                                B3OBxFwn4lgxaNDsNIECQQCTrHWrS/v5rh5sNNPC8dn7e4TThuvQ5l/mv5zTZHu9
                                rtdkt0kn4xawfR5p0nrW6HL2M3pRbJ0obi6h//HZc8KX
                                -----END RSA PRIVATE KEY-----"""
                                  }},
                                  {
                                      "rule_type": "String",
                                      "field": "from",
                                      "params": {
                                          "minimum_length": 5,
                                          "maximum_length": 1024,
                                          "regex": "[^@]+@[^\\.]..*[^\\.]",
                                          "fallback_mode": "remove_record",
                                          "decrypt": True,
                                          "private_key": """-----BEGIN RSA PRIVATE KEY-----
                                MIICXQIBAAKBgQDDmPPm5UC8rXn4uX37m4tN/j4TMAhUVyxN7V7QxMF3HDg5rkl/
                                Ju53DPJbv59TCvlTCXw1ihp9asVyyYpCqrsKCh10sZI0kIrkizlKaB/20Q4P1kYO
                                Cgv4Cwds7Iu2y0TFwDosK9a7MPR9IksL7QRWKjD0DoNemKEpyCt2dZTaQwIDAQAB
                                AoGBAIxvLv8irp5JN3+7Ppb+EMDIwCzqmbBkfmzc9uyRuA4a9suDNUXT3/ceFoyF
                                QpFPGb/i2YlfRLGzjVxjE5/WET9bf//ATOkZyn2sl4jBGs9WIF4by+246E+dfol9
                                /326rSpJjVeklmELU3nJp4ViZzYX+Cwc0rucpZLMqLUUFbvhAkEAzP+WSV2cp5w4
                                E1OBDM7cyq3lneWP0N1OdteZPZlJT0ojsQuoggqvcUMeC/eSoaHyZGN7GS4OVtaC
                                BR++kkPiCwJBAPRCnk4FmeLtnDpTX7KV8a766jSqkDzqEKVVqoGfLYCe0uaWFcwK
                                MKBik0GJ6tFLg9JvJMbP1+eD8+2uQD7rg6kCQDNghyDiBkX3oBIv5nL4UVu2k4qs
                                IwwcuvKL/Er05OurUCCqJFRbKzc+tAQZyzUZKm/AgvR/l3ZqEnIIT7HGs5sCQDej
                                NQvwmqzmEr/2XcYAAZ0p6k80ysYVStVePghoiaTSiJedeDmR2KGv0nsLP0GNQemd
                                B3OBxFwn4lgxaNDsNIECQQCTrHWrS/v5rh5sNNPC8dn7e4TThuvQ5l/mv5zTZHu9
                                rtdkt0kn4xawfR5p0nrW6HL2M3pRbJ0obi6h//HZc8KX
                                -----END RSA PRIVATE KEY-----"""
                                      }},
                                  {
                                      "rule_type": "Date",
                                      "field": "date",
                                      "params": {
                                          "date_format": "%Y-%m-%d %H:%M:%S",
                                          "fallback_mode": "remove_record",
                                          "range_minimum": "1970-01-01 00:00:00",
                                          "range_maximum": "2100-01-01 00:00:00",
                                      }},
                                  {
                                      "rule_type": "String",
                                      "field": "subject",
                                      "params": {
                                          "minimum_length": 0,
                                          "maximum_length": 1024,
                                          "fallback_mode": "remove_record"
                                      }}]
                          },
                          report=[
                              {

                                  'field': 'to',
                                  'rule': 'String',
                                  'number_records': 6,
                                  'category': 'modified',
                                  'description': 'Decrypted all values',
                                  'df': [['steve@apple.com'],
                                         ['wos@apple.com'],
                                         ['billg@microsoft.com'],
                                         ['steve@apple.com'],
                                         ['wos@apple.com'],
                                         ['billg@microsoft.com']],

                              },
                              {

                                  'field': 'from',
                                  'rule': 'String',
                                  'number_records': 6,
                                  'category': 'modified',
                                  'description': 'Decrypted all values',
                                  'df': [['wos@apple.com'],
                                         ['steve@apple.com'],
                                         ['steve@apple.com'],
                                         ['wos@apple.com'],
                                         ['steve@apple.com'],
                                         ['wos@apple.com']],

                              },
                          ])

    def test_email_encryption_inconsistent(self):
        csv = CSVHandler(base_path=self.base_path)

        df = csv.get_dataframe("anonymization/emails.csv")

        scrubber = Scrubber()

        scrubber.run(df=df,
                     config={"rules": [
                         {
                             "rule_type": "String",
                             "field": "to",
                             "params": {
                                 "minimum_length": 5,
                                 "maximum_length": 1024,
                                 "regex": "[^@]+@[^\\.]..*[^\\.]",
                                 "fallback_mode": "remove_record",
                                 "encrypt": True,
                                 "public_key": """-----BEGIN PUBLIC KEY-----
                           MIGfMA0GCSqGSIb3DQEBAQUAA4GNADCBiQKBgQDDmPPm5UC8rXn4uX37m4tN/j4T
                           MAhUVyxN7V7QxMF3HDg5rkl/Ju53DPJbv59TCvlTCXw1ihp9asVyyYpCqrsKCh10
                           sZI0kIrkizlKaB/20Q4P1kYOCgv4Cwds7Iu2y0TFwDosK9a7MPR9IksL7QRWKjD0
                           DoNemKEpyCt2dZTaQwIDAQAB
                           -----END PUBLIC KEY-----"""
                             }}]})

        scrubber.run(df=df,
                     config={"rules": [
                         {
                             "rule_type": "String",
                             "field": "to",
                             "params": {
                                 "minimum_length": 5,
                                 "maximum_length": 1024,
                                 "regex": "[^@]+@[^\\.]..*[^\\.]",
                                 "fallback_mode": "remove_record",
                                 "decrypt": True,
                                 "private_key": """-----BEGIN RSA PRIVATE KEY-----
                            MIICXQIBAAKBgQDDmPPm5UC8rXn4uX37m4tN/j4TMAhUVyxN7V7QxMF3HDg5rkl/
                            Ju53DPJbv59TCvlTCXw1ihp9asVyyYpCqrsKCh10sZI0kIrkizlKaB/20Q4P1kYO
                            Cgv4Cwds7Iu2y0TFwDosK9a7MPR9IksL7QRWKjD0DoNemKEpyCt2dZTaQwIDAQAB
                            AoGBAIxvLv8irp5JN3+7Ppb+EMDIwCzqmbBkfmzc9uyRuA4a9suDNUXT3/ceFoyF
                            QpFPGb/i2YlfRLGzjVxjE5/WET9bf//ATOkZyn2sl4jBGs9WIF4by+246E+dfol9
                            /326rSpJjVeklmELU3nJp4ViZzYX+Cwc0rucpZLMqLUUFbvhAkEAzP+WSV2cp5w4
                            E1OBDM7cyq3lneWP0N1OdteZPZlJT0ojsQuoggqvcUMeC/eSoaHyZGN7GS4OVtaC
                            BR++kkPiCwJBAPRCnk4FmeLtnDpTX7KV8a766jSqkDzqEKVVqoGfLYCe0uaWFcwK
                            MKBik0GJ6tFLg9JvJMbP1+eD8+2uQD7rg6kCQDNghyDiBkX3oBIv5nL4UVu2k4qs
                            IwwcuvKL/Er05OurUCCqJFRbKzc+tAQZyzUZKm/AgvR/l3ZqEnIIT7HGs5sCQDej
                            NQvwmqzmEr/2XcYAAZ0p6k80ysYVStVePghoiaTSiJedeDmR2KGv0nsLP0GNQemd
                            B3OBxFwn4lgxaNDsNIECQQCTrHWrS/v5rh5sNNPC8dn7e4TThuvQ5l/mv5zTZHu9
                            rtdkt0kn4xawfR5p0nrW6HL2M3pRbJ0obi6h//HZc8KX
                            -----END RSA PRIVATE KEY-----"""
                             }}]})

        self._compare_df_to_file(csv, df, "anonymization/emails.csv")



import unittest
import logging
from os import path
from datetime import datetime
import pandas as pd
from dativa.analyzer import DataFrameAnalyzer

logger = logging.getLogger("dativa.analyzer.tests")


class FileAnalyserTests(unittest.TestCase):

    def setUp(self):
        self.base_path = "{0}/test-data/analyzer/".format(
            path.dirname(path.abspath(__file__)))
        super(FileAnalyserTests, self).setUp()

    def test_exception(self):
        logger.debug("testing exception")
        df = pd.read_csv(self.base_path + "decimal_places.csv")

        has_exception = False
        try:
            DataFrameAnalyzer(df)
        except Exception:
            has_exception = True
        self.assertTrue(has_exception)

    def test_decimal_places(self):
        logger.debug("testing decimal places file...")
        df = pd.read_csv(self.base_path +
                         "decimal_places.csv", dtype=pd.np.str)
        analyser = DataFrameAnalyzer(df)
        columns = analyser.columns

        self.assertEqual(columns[0].name, "Number")
        self.assertEqual(columns[0].type, "number")
        self.assertEqual(columns[0].min, -10)
        self.assertEqual(columns[0].max, 10)
        self.assertEqual(columns[0].decimal_places, 1)

        logger.debug("testing no decimal places file...")
        df = pd.read_csv(self.base_path +
                         "no_decimal_places.csv", dtype=pd.np.str)
        analyser = DataFrameAnalyzer(df)
        columns = analyser.columns

        self.assertEqual(columns[0].name, "Number")
        self.assertEqual(columns[0].type, "number")
        self.assertEqual(columns[0].min, 1)
        self.assertEqual(columns[0].max, 10)
        self.assertEqual(columns[0].decimal_places, 0)

    def test_epg_file(self):
        logger.debug("testing EPG file...")
        df = pd.read_csv(self.base_path + "epg-jan-2017.csv", dtype=pd.np.str)
        analyser = DataFrameAnalyzer(df)
        columns = analyser.columns

        self.assertEqual(columns[0].name, "station_name")
        self.assertEqual(columns[0].type, "string")
        self.assertEqual(columns[0].min, 1)
        self.assertEqual(columns[0].max, 100)
        self.assertEqual(columns[0].number_references, 213)

        self.assertEqual(columns[1].name, "airdate")
        self.assertEqual(columns[1].type, "date")
        self.assertEqual(columns[1].date_format, "%Y-%m-%d %H:%M:%S")
        self.assertEqual(columns[1].min, datetime.strptime(
            "2017-01-01 00:00:00", "%Y-%m-%d %H:%M:%S"))
        self.assertEqual(datetime.strptime(
            "2017-02-28 23:59:59.999999", "%Y-%m-%d %H:%M:%S.%f"), columns[1].max)

        self.assertEqual(columns[2].name, "title")
        self.assertEqual(columns[2].type, "string")
        self.assertEqual(columns[2].min, 1)
        self.assertEqual(columns[2].max, 100)

        self.assertEqual(columns[3].name, "epi_num")
        self.assertEqual(columns[3].type, "string")
        self.assertEqual(columns[3].min, 1)
        self.assertGreater(columns[3].max, 9)

        self.assertEqual(columns[4].name, "epi_title")
        self.assertEqual(columns[4].type, "string")
        self.assertEqual(columns[4].min, 1)
        self.assertGreater(columns[4].max, 99)

        self.assertEqual(columns[5].name, "genre")
        self.assertEqual(columns[5].type, "string")
        self.assertGreater(columns[5].min, -1)
        self.assertEqual(columns[5].max, 100)

        self.assertEqual(columns[6].name, "year")
        self.assertEqual(columns[6].type, "number")
        self.assertEqual(columns[6].decimal_places, 0)
        self.assertEqual(columns[6].min, 1000)
        self.assertEqual(columns[6].max, 3000)

        self.assertEqual(columns[7].name, "language")
        self.assertEqual(columns[7].type, "string")
        self.assertEqual(columns[7].min, 0)
        self.assertEqual(columns[7].max, 10)
        self.assertEqual(columns[7].number_references, 2)

        self.assertEqual(columns[8].name, "country_name")
        self.assertEqual(columns[8].type, "string")
        self.assertEqual(columns[8].min, 0)
        self.assertEqual(columns[8].max, 20)
        self.assertEqual(columns[8].number_references, 87)

        self.assertEqual(analyser.get_unique_columns(
            0.95, 5), ['station_name', 'airdate'])

    def test_vod_file(self):
        logger.debug("testing VOD file...")
        df = pd.read_csv(self.base_path + "vod-jun-2017.csv", dtype=pd.np.str)
        analyser = DataFrameAnalyzer(df)

    def test_date_hour_month(self):
        logger.debug("testing date hour month places file...")
        df = pd.read_csv(self.base_path +
                         "dates_hour_month.csv", dtype=pd.np.str)
        analyser = DataFrameAnalyzer(df)
        columns = analyser.columns

        self.assertEqual(columns[0].name, "hours")
        self.assertEqual(columns[0].type, "date")
        self.assertEqual(columns[0].min, datetime.strptime(
            '2017-12-01 12:00:00', "%Y-%m-%d %H:%M:%S"))
        self.assertEqual(columns[0].max, datetime.strptime(
            '2017-12-01 12:59:59.999999', "%Y-%m-%d %H:%M:%S.%f"))

        self.assertEqual(columns[1].name, "days")
        self.assertEqual(columns[1].type, "date")
        self.assertEqual(columns[1].min, datetime.strptime(
            '2017-12-01 00:00:00', "%Y-%m-%d %H:%M:%S"))
        self.assertEqual(columns[1].max, datetime.strptime(
            '2017-12-31 23:59:59.999999', "%Y-%m-%d %H:%M:%S.%f"))



# (c) 2012-2018 Dativa, all rights reserved
# -----------------------------------------
# Usage subject to license agreement
# hello@dativa.com for more information

import pandas as pd
from dativa.scrubber.tests import _BaseTest
from dativa.tools.pandas import CSVHandler
from dativa.scrubber import Scrubber, ScrubberValidationError

class DateTests(_BaseTest):

    def test_date(self):
        self._test_filter(dirty_file="date/date_format.csv",
                          clean_file="date/date_format_clean.csv",
                          config={
                              "rules": [
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": False,
                                          "date_format": "%d-%m-%Y %H:%M:%S",
                                          "default_value": "N/A",
                                          "fallback_mode": "use_default",
                                          "is_unique": False,
                                          "lookalike_match": False,
                                          "range_check": "rolling",
                                          "range_maximum": "10000",
                                          "range_minimum": "0",
                                          "skip_blank": False,
                                          "string_distance_threshold": 0.7
                                      },
                                      "rule_type": "Date",
                                      "field": "Date"
                                  }
                              ]
                          },
                          report=[
                              {

                                  'field': 'Date',
                                  'rule': 'Date',
                                  'number_records': 3,
                                  'category': 'replaced',
                                  'description': 'Replaced with default value',
                                  'df': [['30-03-2010 10:03:26', 'N/A'],
                                         ['31-02-2027 10:99:16', 'N/A'],
                                         ['05-04-2027 25:03:16', 'N/A']],

                              },
                          ])

    def test_date_1(self):
        self._test_filter(dirty_file="date/date_format.csv",
                          clean_file="date/date_format_clean1.csv",
                          config={
                              "rules": [
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": False,
                                          "date_format": "%d-%m-%Y %H:%M:%S",
                                          "default_value": "",
                                          "fallback_mode": "remove_record",

                                          "is_unique": False,
                                          "lookalike_match": False,
                                          "range_check": "fixed",
                                          "range_maximum": "01-01-2028 00:00:00",
                                          "range_minimum": "01-01-2017 10:03:16",
                                          "skip_blank": False,
                                          "string_distance_threshold": 0.7
                                      },
                                      "rule_type": "Date",
                                      "field": "Date"
                                  }
                              ]
                          },
                          report=[
                              {

                                  'field': 'Date',
                                  'rule': 'Date',
                                  'number_records': 3,
                                  'category': 'quarantined',
                                  'description': 'Data quarantined',
                                  'df': [['30-03-2010 10:03:26'],
                                         ['31-02-2027 10:99:16'],
                                         ['05-04-2027 25:03:16']],

                              },
                          ])

    def test_date_2(self):
        self._test_filter(dirty_file="date/date_format.csv",
                          clean_file="date/date_format_clean2.csv",
                          config={
                              "rules": [
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": False,
                                          "date_format": "%d-%m-%Y %H:%M:%S",
                                          "default_value": "",
                                          "fallback_mode": "do_not_replace",

                                          "is_unique": False,
                                          "lookalike_match": False,
                                          "range_check": "none",
                                          "range_maximum": "",
                                          "range_minimum": "",
                                          "skip_blank": False,
                                          "string_distance_threshold": 0.7
                                      },
                                      "rule_type": "Date",
                                      "field": "Date"
                                  }
                              ]
                          },
                          report=[
                              {

                                  'field': 'Date',
                                  'rule': 'Date',
                                  'number_records': 2,
                                  'category': 'ignored',
                                  'description': 'No changes made',
                                  'df': [['31-02-2027 10:99:16', '31-02-2027 10:99:16'],
                                         ['05-04-2027 25:03:16', '05-04-2027 25:03:16']],

                              },
                          ])

    def test_date_3(self):
        self._test_filter(dirty_file="date/date_format1.csv",
                          clean_file="date/date_format1_clean.csv",
                          config={
                              "rules": [
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": False,
                                          "date_format": "%d-%m-%Y",
                                          "default_value": "N/A",
                                          "fallback_mode": "remove_record",

                                          "is_unique": False,
                                          "lookalike_match": False,
                                          "range_check": "fixed",
                                          "range_maximum": "02-05-2021",
                                          "range_minimum": "01-01-2001",
                                          "skip_blank": False,
                                          "string_distance_threshold": 0.7
                                      },
                                      "rule_type": "Date",
                                      "field": "Date"
                                  }
                              ]
                          },
                          report=[
                              {

                                  'field': 'Date',
                                  'rule': 'Date',
                                  'number_records': 10,
                                  'category': 'quarantined',
                                  'description': 'Data quarantined',
                                  'df': [['19-19-2001'],
                                         ['05-04-2103'],
                                         ['06-04-2103'],
                                         ['07-04-2103'],
                                         ['08-04-2103'],
                                         ['09-04-2103'],
                                         [None],
                                         ['748949+84'],
                                         ['231514'],
                                         ['45457']],

                              },
                          ]
                          )

    def test_date_4(self):
        self._test_filter(dirty_file="date/date_is_unique_dirty.csv",
                          clean_file="date/date_is_unique_cleaned.csv",
                          config={
                              "rules": [
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": False,
                                          "date_format": "%d/%m/%Y %H:%M:%S",
                                          "default_value": "N/A",
                                          "fallback_mode": "use_default",

                                          "is_unique": True,
                                          "lookalike_match": False,
                                          "range_check": "fixed",
                                          "range_maximum": "24/12/2017 07:01:15",
                                          "range_minimum": "18/12/2017 07:01:15",
                                          "skip_blank": False,
                                          "string_distance_threshold": 0.7
                                      },
                                      "rule_type": "Date",
                                      "field": "Date"
                                  }
                              ]
                          },
                          report=[
                              {

                                  'field': 'Date',
                                  'rule': 'Date',
                                  'number_records': 2,
                                  'category': 'quarantined',
                                  'description': 'Removed duplicates',
                                  'df': [['18/12/2017 07:01:15'],
                                         ['19/12/2017 07:01:15']],

                              },
                              {

                                  'field': 'Date',
                                  'rule': 'Date',
                                  'number_records': 1,
                                  'category': 'replaced',
                                  'description': 'Replaced with default value',
                                  'df': [['25/12/2017 07:01:15', 'N/A']],

                              },
                          ])

    def test_date_5(self):
        self._test_filter(dirty_file="date/date_format_2.csv",
                          clean_file="date/date_format_2.csv",
                          config={'rules': [
                              {'field': 'PREV_CHECKIN_DATE',
                               'rule_type': 'Date',
                               'params': {'date_format': '%Y-%m-%d %H:%M:%S.%f', 'range_check': 'none',
                                          'fallback_mode': 'remove_record'}},
                              {'field': 'CHECKIN_DATE',
                               'rule_type': 'Date',
                               'params': {'date_format': '%Y-%m-%d %H:%M:%S.%f', 'range_check': 'none',
                                          'fallback_mode': 'remove_record'}}, ]},
                          csv_delimiter="|",
                          report=[])

    def test_date_6(self):
        self._test_filter(dirty_file="date/date_format_2.csv",
                          clean_file="date/date_format_2_clean.csv",
                          config={'rules': [
                              {'field': 'PREV_CHECKIN_DATE',
                               'rule_type': 'Date',
                               'params': {'date_format': '%Y/%m/%d %H:%M:%S.%f', 'range_check': 'none',
                                          'fallback_mode': 'remove_record'}},
                              {'field': 'CHECKIN_DATE',
                               'rule_type': 'Date',
                               'params': {'date_format': '%Y/%m/%d %H:%M:%S.%f', 'range_check': 'none',
                                          'fallback_mode': 'remove_record'}}, ]},
                          csv_delimiter="|",
                          report=[
                              {

                                  'field': 'PREV_CHECKIN_DATE',
                                  'rule': 'Date',
                                  'number_records': 11,
                                  'category': 'quarantined',
                                  'description': 'Data quarantined',
                                  'df': [['2018-02-28 00:00:00.000000'],
                                         ['2018-02-28 00:00:00.000000'],
                                         ['2018-02-25 00:00:00.000000'],
                                         ['2018-02-28 00:00:00.000000'],
                                         ['2018-02-27 00:00:00.000000'],
                                         ['2018-02-24 00:00:00.000000'],
                                         ['2018-02-28 00:00:00.000000'],
                                         ['2018-02-28 00:00:00.000000'],
                                         ['2018-02-28 00:00:00.000000'],
                                         ['2018-02-28 00:00:00.000000'],
                                         ['2018-02-28 00:00:00.000000']],

                              },
                          ])

    def test_date_epoch(self):
        self._test_filter(dirty_file="date/check_epoch_time.csv",
                          clean_file="date/check_epoch_time_cleaned1.csv",
                          config={
                              "rules": [
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": False,
                                          "date_format": "%s",
                                          "default_value": "N/A",
                                          "fallback_mode": "use_default",

                                          "is_unique": True,
                                          "lookalike_match": False,
                                          "range_check": "fixed",
                                          "range_maximum": "1511373354",
                                          "range_minimum": "1511344324",
                                          "skip_blank": False,
                                          "string_distance_threshold": 0.7
                                      },
                                      "rule_type": "Date",
                                      "field": "utc_air_start_epoch"
                                  }
                              ]
                          },
                          report=[
                              {

                                  'field': 'utc_air_start_epoch',
                                  'rule': 'Date',
                                  'number_records': 14,
                                  'category': 'replaced',
                                  'description': 'Replaced with default value',
                                  'df': [[1511379769, 'N/A'],
                                         [1511343124, 'N/A'],
                                         [1511336569, 'N/A'],
                                         [1511329924, 'N/A'],
                                         [1511328724, 'N/A'],
                                         [1511322169, 'N/A'],
                                         [1511315524, 'N/A'],
                                         [1511314324, 'N/A'],
                                         [1511394199, 'N/A'],
                                         [1511393209, 'N/A'],
                                         [1511386999, 'N/A'],
                                         [1511386009, 'N/A'],
                                         [1511379799, 'N/A'],
                                         [1511378809, 'N/A']],

                              },
                          ])

    def test_date_epoch_2(self):
        self._test_filter(dirty_file="date/check_epoch_time.csv",
                          clean_file="date/check_epoch_time_cleaned1.csv",
                          config={
                              "rules": [
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": False,
                                          "date_format": "%s",
                                          "default_value": "N/A",
                                          "fallback_mode": "use_default",

                                          "is_unique": True,
                                          "lookalike_match": False,
                                          "range_check": "fixed",
                                          "range_maximum": "1511373354",
                                          "range_minimum": "banana",
                                          "skip_blank": False,
                                          "string_distance_threshold": 0.7
                                      },
                                      "rule_type": "Date",
                                      "field": "utc_air_start_epoch"
                                  }
                              ]
                          },
                          report=[
                              {

                                  'field': 'Date',
                                  'rule': 'Date',
                                  'number_records': 3,
                                  'category': 'replaced',
                                  'description': 'Replaced with default value',
                                  'df': [['30-03-2010 10:03:26', 'N/A'],
                                         ['31-02-2027 10:99:16', 'N/A'],
                                         ['05-04-2027 25:03:16', 'N/A']],

                              },
                          ],
                          expected_error=ScrubberValidationError)

    def test_date_range_error(self):
        self._test_filter(dirty_file="date/check_epoch_time.csv",
                          clean_file="date/check_epoch_time_cleaned1.csv",
                          config={
                              "rules": [
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": False,
                                          "date_format": "%s",
                                          "default_value": "N/A",
                                          "fallback_mode": "use_default",

                                          "is_unique": True,
                                          "lookalike_match": False,
                                          "range_check": "fixed",
                                          "range_minimum": "1511344324",
                                          "skip_blank": False,
                                          "string_distance_threshold": 0.7
                                      },
                                      "rule_type": "Date",
                                      "field": "utc_air_start_epoch"
                                  }
                              ]
                          },
                          report=[
                              {

                                  'field': 'Date',
                                  'rule': 'Date',
                                  'number_records': 3,
                                  'category': 'replaced',
                                  'description': 'Replaced with default value',
                                  'df': [['30-03-2010 10:03:26', 'N/A'],
                                         ['31-02-2027 10:99:16', 'N/A'],
                                         ['05-04-2027 25:03:16', 'N/A']],

                              },
                          ],
                          expected_error=ScrubberValidationError)

    def test_date_range_error_2(self):
        self._test_filter(dirty_file="date/check_epoch_time.csv",
                          clean_file="date/check_epoch_time_cleaned1.csv",
                          config={
                              "rules": [
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": False,
                                          "date_format": "%s",
                                          "default_value": "N/A",
                                          "fallback_mode": "use_default",

                                          "is_unique": True,
                                          "lookalike_match": False,
                                          "range_check": "fixed",
                                          "range_maximum": "1511373354",
                                          "skip_blank": False,
                                          "string_distance_threshold": 0.7
                                      },
                                      "rule_type": "Date",
                                      "field": "utc_air_start_epoch"
                                  }
                              ]
                          },
                          report=[
                              {

                                  'field': 'Date',
                                  'rule': 'Date',
                                  'number_records': 3,
                                  'category': 'replaced',
                                  'description': 'Replaced with default value',
                                  'df': [['30-03-2010 10:03:26', 'N/A'],
                                         ['31-02-2027 10:99:16', 'N/A'],
                                         ['05-04-2027 25:03:16', 'N/A']],

                              },
                          ],
                          expected_error=ScrubberValidationError)

    def test_date_range_error_3(self):
        self._test_filter(dirty_file="date/check_epoch_time.csv",
                          clean_file="date/check_epoch_time_cleaned1.csv",
                          config={
                              "rules": [
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": False,
                                          "date_format": "%s",
                                          "default_value": "N/A",
                                          "fallback_mode": "use_default",

                                          "is_unique": True,
                                          "lookalike_match": False,
                                          "range_check": "fixed",
                                          "range_maximum": "1st January 2001",
                                          "range_minimum": "1511344324",
                                          "skip_blank": False,
                                          "string_distance_threshold": 0.7
                                      },
                                      "rule_type": "Date",
                                      "field": "utc_air_start_epoch"
                                  }
                              ]
                          },
                          report=[
                              {

                                  'field': 'Date',
                                  'rule': 'Date',
                                  'number_records': 3,
                                  'category': 'replaced',
                                  'description': 'Replaced with default value',
                                  'df': [['30-03-2010 10:03:26', 'N/A'],
                                         ['31-02-2027 10:99:16', 'N/A'],
                                         ['05-04-2027 25:03:16', 'N/A']],

                              },
                          ],
                          expected_error=ScrubberValidationError)

    def test_date_range_error_4(self):
        self._test_filter(dirty_file="date/check_epoch_time.csv",
                          clean_file="date/check_epoch_time_cleaned1.csv",
                          config={
                              "rules": [
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": False,
                                          "date_format": "%s",
                                          "default_value": "N/A",
                                          "fallback_mode": "use_default",

                                          "is_unique": True,
                                          "lookalike_match": False,
                                          "range_check": "fixed",
                                          "range_maximum": "1511373354",
                                          "range_minimum": "1st January 2001",
                                          "skip_blank": False,
                                          "string_distance_threshold": 0.7
                                      },
                                      "rule_type": "Date",
                                      "field": "utc_air_start_epoch"
                                  }
                              ]
                          },
                          report=[
                              {

                                  'field': 'Date',
                                  'rule': 'Date',
                                  'number_records': 3,
                                  'category': 'replaced',
                                  'description': 'Replaced with default value',
                                  'df': [['30-03-2010 10:03:26', 'N/A'],
                                         ['31-02-2027 10:99:16', 'N/A'],
                                         ['05-04-2027 25:03:16', 'N/A']],

                              },
                          ],
                          expected_error=ScrubberValidationError)

    def test_date_range_error_5(self):
        self._test_filter(dirty_file="date/check_epoch_time.csv",
                          clean_file="date/check_epoch_time_cleaned1.csv",
                          config={
                              "rules": [
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": False,
                                          "date_format": "%s",
                                          "default_value": "N/A",
                                          "fallback_mode": "use_default",

                                          "is_unique": True,
                                          "lookalike_match": False,
                                          "range_check": "fixed",
                                          "range_maximum": "1511373354",
                                          "range_minimum": "2511344324",
                                          "skip_blank": False,
                                          "string_distance_threshold": 0.7
                                      },
                                      "rule_type": "Date",
                                      "field": "utc_air_start_epoch"
                                  }
                              ]
                          },
                          report=[
                              {

                                  'field': 'Date',
                                  'rule': 'Date',
                                  'number_records': 3,
                                  'category': 'replaced',
                                  'description': 'Replaced with default value',
                                  'df': [['30-03-2010 10:03:26', 'N/A'],
                                         ['31-02-2027 10:99:16', 'N/A'],
                                         ['05-04-2027 25:03:16', 'N/A']],

                              },
                          ],
                          expected_error=ScrubberValidationError)

    def test_date_error_6(self):
        self._test_filter(dirty_file="date/date_format.csv",
                          clean_file="date/date_format_clean.csv",
                          config={
                              "rules": [
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": False,
                                          "date_format": "%d-%m-%Y %H:%M:%S",
                                          "default_value": "N/A",
                                          "fallback_mode": "use_default",
                                          "is_unique": False,
                                          "lookalike_match": False,
                                          "range_check": "rolling",
                                          "range_maximum": "10000",
                                          "range_minimum": "1st January 2000",
                                          "skip_blank": False,
                                          "string_distance_threshold": 0.7
                                      },
                                      "rule_type": "Date",
                                      "field": "Date"
                                  }
                              ]
                          },
                          report=[
                              {

                                  'field': 'Date',
                                  'rule': 'Date',
                                  'number_records': 3,
                                  'category': 'replaced',
                                  'description': 'Replaced with default value',
                                  'df': [['30-03-2010 10:03:26', 'N/A'],
                                         ['31-02-2027 10:99:16', 'N/A'],
                                         ['05-04-2027 25:03:16', 'N/A']],

                              },
                          ],
                          expected_error=ScrubberValidationError)

    def test_date_error_7(self):
        self._test_filter(dirty_file="date/date_format.csv",
                          clean_file="date/date_format_clean.csv",
                          config={
                              "rules": [
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": False,
                                          "date_format": "%d-%m-%Y %H:%M:%S",
                                          "default_value": "N/A",
                                          "fallback_mode": "use_default",
                                          "is_unique": False,
                                          "lookalike_match": False,
                                          "range_check": "rolling",
                                          "range_maximum": "1st January 2000",
                                          "range_minimum": "0",
                                          "skip_blank": False,
                                          "string_distance_threshold": 0.7
                                      },
                                      "rule_type": "Date",
                                      "field": "Date"
                                  }
                              ]
                          },
                          report=[
                              {

                                  'field': 'Date',
                                  'rule': 'Date',
                                  'number_records': 3,
                                  'category': 'replaced',
                                  'description': 'Replaced with default value',
                                  'df': [['30-03-2010 10:03:26', 'N/A'],
                                         ['31-02-2027 10:99:16', 'N/A'],
                                         ['05-04-2027 25:03:16', 'N/A']],

                              },
                          ],
                          expected_error=ScrubberValidationError)

    def test_date_as_timestamp(self):
        csv = CSVHandler(base_path=self.base_path)

        df = csv.get_dataframe("anonymization/emails.csv")

        df.date = pd.to_datetime(df.date)

        scrubber = Scrubber()

        scrubber.run(df=df,
                     config={"rules": [
                         {
                             "rule_type": "String",
                             "field": "to",
                             "params": {
                                 "minimum_length": 5,
                                 "maximum_length": 1024,
                                 "regex": "[^@]+@[^\\.]..*[^\\.]",
                                 "fallback_mode": "remove_record",
                                 "encrypt": True,
                                 "public_key": """-----BEGIN PUBLIC KEY-----
                           MIGfMA0GCSqGSIb3DQEBAQUAA4GNADCBiQKBgQDDmPPm5UC8rXn4uX37m4tN/j4T
                           MAhUVyxN7V7QxMF3HDg5rkl/Ju53DPJbv59TCvlTCXw1ihp9asVyyYpCqrsKCh10
                           sZI0kIrkizlKaB/20Q4P1kYOCgv4Cwds7Iu2y0TFwDosK9a7MPR9IksL7QRWKjD0
                           DoNemKEpyCt2dZTaQwIDAQAB
                           -----END PUBLIC KEY-----"""
                             }},
                         {
                             "rule_type": "Date",
                             "field": "date",
                             "params": {
                                 "date_format": "%Y-%m-%d %H:%M:%S",
                                 "fallback_mode": "remove_record"
                             }
                         }]})

        scrubber.run(df=df,
                     config={"rules": [
                         {
                             "rule_type": "String",
                             "field": "to",
                             "params": {
                                 "minimum_length": 5,
                                 "maximum_length": 1024,
                                 "regex": "[^@]+@[^\\.]..*[^\\.]",
                                 "fallback_mode": "remove_record",
                                 "decrypt": True,
                                 "private_key": """-----BEGIN RSA PRIVATE KEY-----
                            MIICXQIBAAKBgQDDmPPm5UC8rXn4uX37m4tN/j4TMAhUVyxN7V7QxMF3HDg5rkl/
                            Ju53DPJbv59TCvlTCXw1ihp9asVyyYpCqrsKCh10sZI0kIrkizlKaB/20Q4P1kYO
                            Cgv4Cwds7Iu2y0TFwDosK9a7MPR9IksL7QRWKjD0DoNemKEpyCt2dZTaQwIDAQAB
                            AoGBAIxvLv8irp5JN3+7Ppb+EMDIwCzqmbBkfmzc9uyRuA4a9suDNUXT3/ceFoyF
                            QpFPGb/i2YlfRLGzjVxjE5/WET9bf//ATOkZyn2sl4jBGs9WIF4by+246E+dfol9
                            /326rSpJjVeklmELU3nJp4ViZzYX+Cwc0rucpZLMqLUUFbvhAkEAzP+WSV2cp5w4
                            E1OBDM7cyq3lneWP0N1OdteZPZlJT0ojsQuoggqvcUMeC/eSoaHyZGN7GS4OVtaC
                            BR++kkPiCwJBAPRCnk4FmeLtnDpTX7KV8a766jSqkDzqEKVVqoGfLYCe0uaWFcwK
                            MKBik0GJ6tFLg9JvJMbP1+eD8+2uQD7rg6kCQDNghyDiBkX3oBIv5nL4UVu2k4qs
                            IwwcuvKL/Er05OurUCCqJFRbKzc+tAQZyzUZKm/AgvR/l3ZqEnIIT7HGs5sCQDej
                            NQvwmqzmEr/2XcYAAZ0p6k80ysYVStVePghoiaTSiJedeDmR2KGv0nsLP0GNQemd
                            B3OBxFwn4lgxaNDsNIECQQCTrHWrS/v5rh5sNNPC8dn7e4TThuvQ5l/mv5zTZHu9
                            rtdkt0kn4xawfR5p0nrW6HL2M3pRbJ0obi6h//HZc8KX
                            -----END RSA PRIVATE KEY-----"""
                             }}]})

        self._compare_df_to_file(csv, df, "anonymization/emails.csv")



import logging
# (c) 2012-2018 Dativa, all rights reserved
# -----------------------------------------
#  This code is licensed under MIT license (see license.txt for details)

import re
import pandas as pd
from dativa.analyzer.analyzer import DataFrameAnalyzer
from dativa.scrubber.base import DateRangeType, FallbackType, SessionGapsType, SessionOverlapsType, DefaultProfile
from dativa.scrubber.validators import StringValidator, NumberValidator, DateValidator, LookupValidator, \
    UniqueFieldsValidator, SessionValidator
from dativa.tools.pandas import datetime_to_string

logger = logging.getLogger("dativa.analyzer.autoconfig")


class FileCounter:

    def __init__(self):
        self.df_dict = {}
        self.profile = DefaultProfile()
        self.maximum_string_length = 1024

    def clear_files(self):
        self.df_dict = {}

    def add_df(self, column, list_of_values):
        ix = "{0}.ref.csv".format("".join(c if c.isalnum() or c in ('.', '_') else '-' for c in column.name).strip())
        self.df_dict[ix] = pd.DataFrame(list_of_values, columns=[column.name])
        return ix


class AutoConfig:

    def __init__(self):
        self.files = FileCounter()
        self.maximum_string_length = 1024
        self.profile = DefaultProfile()

    @staticmethod
    def _get_attr(params, attr, default=None):
        if attr in params:
            return params[attr]
        else:
            if default is None:
                return ""
            else:
                return default

    @staticmethod
    def _get_fallback_mode(column):
        if column.fallback_mode == 'quarantine':
            return FallbackType.REMOVE_ENTRY
        elif column.fallback_mode == 'default':
            return FallbackType.USE_DEFAULT_VALUE
        elif column.fallback_mode == 'closest':
            return FallbackType.REMOVE_ENTRY

    def _trim_string(self, s):
        return s[0: self.maximum_string_length]

    @staticmethod
    def _finalise_params(validator, params, default_value=""):
        # remove any default values
        for a in list(params):
            if params[a] == validator.fields[a]["default"]:
                del params[a]

        # add the default for that fallback mode
        if "fallback_mode" in params and params["fallback_mode"] == "use_default":
            params["default_value"] = default_value

        return (params)

    def _get_string_params(self, column, max_references):
        regex = ".*"
        if 1 <= column.number_references <= max_references:
            regex = ""
            for r in column.reference_list:
                regex = regex + re.escape(r) + "|"
            regex = regex[:-1]
        return self._finalise_params(
            validator=StringValidator(),
            params={"minimum_length": column.min,
                    "maximum_length": column.max,
                    "attempt_closest_match": False,
                    "lookalike_match": False,
                    "fallback_mode": self._get_fallback_mode(column),
                    "skip_blank": column.has_blanks,
                    "is_unique": column.is_unique,
                    "regex": regex},

            default_value=self._trim_string(column.default_value))

    def _get_number_params(self, column):
        return self._finalise_params(
            validator=NumberValidator(),
            params={"decimal_places": column.decimal_places,
                    "fix_decimal_places": True,
                    "minimum_value": column.min,
                    "maximum_value": column.max,
                    "attempt_closest_match": False,
                    "lookalike_match": False,
                    "fallback_mode": self._get_fallback_mode(column),
                    "skip_blank": column.has_blanks,
                    "is_unique": column.is_unique},
            default_value=self._trim_string(column.default_value))

    def _get_date_params(self, column):
        return self._finalise_params(
            validator=DateValidator(),
            params={"date_format": column.date_format,
                    "range_check": DateRangeType.FIXED,
                    "range_minimum": datetime_to_string(column.min, column.date_format),
                    "range_maximum": datetime_to_string(column.max, column.date_format),
                    "attempt_closest_match": False,
                    "lookalike_match": False,
                    "fallback_mode": self._get_fallback_mode(column),
                    "skip_blank": column.has_blanks,
                    "is_unique": column.is_unique},
            default_value=self._trim_string(column.default_value))

    def _get_lookup_params(self, column):

        file = self.files.add_df(column, column.reference_list)

        attempt_closest_match = (column.fallback_mode == "closest")
        lookalike_match = (column.number_references < 50 and
                           column.sample_percent_blank * column.total_records < self.profile.maximum_records_lookalike)
        return self._finalise_params(
            validator=LookupValidator(),
            params={"original_reference": file,
                    "reference_field": 0,
                    "attempt_closest_match": attempt_closest_match,
                    "lookalike_match": lookalike_match,
                    "fallback_mode": self._get_fallback_mode(column),
                    "skip_blank": column.has_blanks},
            default_value=self._trim_string(column.default_value))

    def _get_rule(self, column, min_references, max_references):
        if column.type == 'string':
            if min_references <= column.number_references <= max_references:
                return {
                    "field": column.name,
                    "rule_type": "Lookup",
                    "params": self._get_lookup_params(column)
                }
            else:
                return {
                    "field": column.name,
                    "rule_type": "String",
                    "params": self._get_string_params(column, max_references)
                }

        if column.type == 'number':
            return {
                "field": column.name,
                "rule_type": "Number",
                "params": self._get_number_params(column)
            }

        if column.type == 'date':
            return {
                "field": column.name,
                "rule_type": "Date",
                "params": self._get_date_params(column)
            }

    @staticmethod
    def _concat_list(l, sep=","):
        s = ""
        for item in l:
            if item == l[-1]:
                return s + item
            else:
                s = s + item + sep

    def _get_unique_rule(self, analyser):
        cols = analyser.get_unique_columns(0.95, 5)
        if cols:
            unique_columns = self._concat_list(cols, ",")
            logger.info(
                "Added uniqueness rule to config")

            return {
                "rule_type": "Uniqueness",
                "params": self._finalise_params(validator=UniqueFieldsValidator(),
                                                params={
                                                    "unique_fields": unique_columns,
                                                    "use_last_value": True
                                                })
            }
        else:
            return None

    @staticmethod
    def _session_rule_is_good(df, session_params):
        number_rows = df.shape[0]

        number_clean_rows = SessionValidator(
            session_params).count_good_rows(df)

        logger.info("Session rule {0}, {1}, {2} has {3} good records".format(session_params["key_field"],
                                                                             session_params["start_field"],
                                                                             session_params["end_field"],
                                                                             number_clean_rows / number_rows * 100))
        if number_clean_rows > number_rows * 0.8:
            return True

        return False

    def _get_session_rule(self, df, rules):
        if len(rules) > 1:
            # more than one session rule so look for same formats
            candidate_columns = []
            for first_rule in rules:
                for second_rule in rules:
                    if first_rule["field"] != second_rule["field"] and first_rule["params"]["date_format"] == \
                            second_rule["params"]["date_format"]:
                        candidate_columns.append([first_rule, second_rule])

            for setting in [{'gaps': SessionGapsType.EXTEND_END, 'allowed_gap': 0},
                            {'gaps': SessionGapsType.EXTEND_END, 'allowed_gap': 1},
                            {'gaps': SessionGapsType.IGNORE, 'allowed_gap': 0}]:
                for candidate in candidate_columns:
                    for key in df.columns:
                        if key != candidate[0]["field"] and key != candidate[1]["field"]:
                            session_params = self._finalise_params(validator=SessionValidator(),
                                                                   params={"key_field": key,
                                                                           "start_field": candidate[0]["field"],
                                                                           "end_field": candidate[1]["field"],
                                                                           "allowed_gap_seconds": 0,
                                                                           "allowed_overlap_seconds": setting[
                                                                               'allowed_gap'],
                                                                           "gaps_option": setting['gaps'],
                                                                           "overlaps_option": SessionOverlapsType.TRUNCATE_END,
                                                                           "date_format": candidate[0]["params"][
                                                                               "date_format"],
                                                                           "remove_zero_length": True})

                            if self._session_rule_is_good(df, session_params):
                                logger.info("Added session rule to config")
                                return {"rule_type": "Session",
                                        "params": session_params}
        return None

    def create_config(self,
                      df,
                      profile=None,
                      csv_delimiter=',',
                      maximum_string_length=1024,
                      large_sample_size=5000,
                      small_sample_size=100,
                      clean_threshold=0.95,
                      outlier_threshold=2,
                      min_occurences=5,
                      min_references=10,
                      max_references=1000):

        self.files.clear_files()
        if profile is not None:
            self.profile = profile
        self.maximum_string_length = maximum_string_length

        """Auto config create rules."""

        # delete any existing rules.
        config = {"rules": []}

        # analyse the dataframe
        analyser = DataFrameAnalyzer(df=df,
                                     large_sample_size=large_sample_size,
                                     small_sample_size=small_sample_size,
                                     clean_threshold=clean_threshold,
                                     outlier_threshold=outlier_threshold,
                                     min_occurences=min_occurences)

        # create rules appropriate for each column
        date_rules = []
        has_unique = False
        for column in analyser.columns:
            has_unique = has_unique or column.is_unique
            rule = self._get_rule(column, min_references, max_references)
            logger.info("Added {0} rule to config".format(rule["rule_type"]))
            config["rules"].append(rule)

            if rule["rule_type"] == 'Date':
                date_rules.append(rule)

        # create unique columns
        if not has_unique:
            unique_rule = self._get_unique_rule(analyser)
            if unique_rule is not None:
                config["rules"].append(unique_rule)

        # add session rules...
        session_rule = self._get_session_rule(df, date_rules)
        if session_rule is not None:
            config["rules"].append(session_rule)

        return config, self.files.df_dict

    def describe_dataframe(self, config, df_dict):

        s = "column,type,minimum,maximum,format,has_blanks\n"
        extra = "\n"
        for rule in config["rules"]:
            params = self._get_attr(rule, "params")
            column = self._get_attr(rule, "field")
            rule_type = self._get_attr(rule, "rule_type")
            if rule_type == "Date":
                s = s + "{column},{type},{minimum},{maximum},{format},{has_blanks}\n".format(
                    column=column,
                    type=rule_type,
                    minimum=self._get_attr(params, "range_minimum"),
                    maximum=self._get_attr(params, "range_maximum"),
                    format=self._get_attr(params, "date_format"),
                    has_blanks=self._get_attr(params, "skip_blank", False))

            elif rule_type == "String":
                s = s + "{column},{type},{minimum},{maximum},{format},{has_blanks}\n".format(
                    column=column,
                    type=rule_type,
                    minimum=self._get_attr(params, "minimum_length"),
                    maximum=self._get_attr(params, "maximum_length"),
                    format='',
                    has_blanks=self._get_attr(params, "skip_blank", False))

            elif rule_type == "Number":
                if self._get_attr(params, "decimal_places") == 0:
                    frm = "int"
                else:
                    frm = "0." + "0" * self._get_attr(params, "decimal_places")

                s = s + "{column},{type},{minimum},{maximum},{format},{has_blanks}\n".format(
                    column=column,
                    type=rule_type,
                    minimum=self._get_attr(params, "minimum_value"),
                    maximum=self._get_attr(params, "maximum_value"),
                    format=frm,
                    has_blanks=self._get_attr(params, "skip_blank", False))

            elif rule_type == "Lookup":
                s = s + "{column},{type},{minimum},{maximum},{format},{has_blanks}\n".format(
                    column=column,
                    type="String",
                    minimum=df_dict[params["original_reference"]
                            ].iloc[:, 0].min().replace(",", "_"),
                    maximum=df_dict[params["original_reference"]
                            ].iloc[:, 0].max().replace(",", "_"),
                    format="Lookup",
                    has_blanks=self._get_attr(params, "skip_blank", False))

            elif rule_type == "Uniqueness":
                extra = extra + "Unique on {uniques}\n".format(
                    uniques=self._get_attr(params, "unique_fields"))

            elif rule_type == "Session":
                extra = extra + "Session rule on {start}, {stop} with unique id {id}\n".format(
                    start=self._get_attr(params, "start_field"),
                    stop=self._get_attr(params, "end_field"),
                    id=self._get_attr(params, "key_field"))

        return s + extra
